"""
Serializer for votes data.

Serializes the votes field in the ContentSerializer.
"""

from typing import Any

from rest_framework import serializers


class VotesSerializer(serializers.Serializer[dict[str, Any]]):
    """
    Serializer for votes data.

    Handles data of type dict[str, int].

    Attributes:
        up (list[str]): List of user IDs who upvoted the content.
        down (list[str]): List of user IDs who downvoted the content.
        up_count (int): Total number of upvotes.
        down_count (int): Total number of downvotes.
        count (int): Total number of votes.
        point (int): The point value of the content.
    """

    up = serializers.ListField(child=serializers.CharField(), default=[])
    down = serializers.ListField(child=serializers.CharField(), default=[])
    up_count = serializers.IntegerField()
    down_count = serializers.IntegerField()
    count = serializers.IntegerField()
    point = serializers.IntegerField()

    def create(self, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError

    def update(self, instance: Any, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError


class VotesInputSerializer(serializers.Serializer[dict[str, Any]]):
    """
    Serializer for handling votes on a content item.

    Attributes:
        user_id (str): The ID of the user casting the vote.
        value (str): The value of the vote, either "up" or "down".
    """

    user_id = serializers.CharField(required=True)
    value = serializers.ChoiceField(choices=["up", "down"], required=True)

    def create(self, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError

    def update(self, instance: Any, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError


class VoteSummarySerializer(serializers.Serializer[dict[str, Any]]):
    """
    Serializer for summarizing votes on a content item.

    Attributes:
        count (int): The total number of votes.
        up_count (int): The number of upvotes.
        down_count (int): The number of downvotes.
        point (int): The net score (upvotes minus downvotes).
    """

    count = serializers.IntegerField(min_value=0)
    up_count = serializers.IntegerField(min_value=0)
    down_count = serializers.IntegerField(min_value=0)
    point = serializers.IntegerField()

    def create(self, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError

    def update(self, instance: Any, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError



"""
Base abstract classes for search features.

To implement a new search backend, one must implement all the methods both from
BaseDocumentSearchBackend, BaseIndexSearchBackend and BaseThreadSearchBackend. Then, the
SearchBackend.*_CLASS class attributes must be set to the corresponding child classes.
"""

import typing as t


class BaseDocumentSearchBackend:
    """
    Abstract base class for document management.
    """

    def index_document(
        self, index_name: str, doc_id: str | int, document: dict[str, t.Any]
    ) -> None:
        raise NotImplementedError

    def update_document(
        self, index_name: str, doc_id: str | int, update_data: dict[str, t.Any]
    ) -> None:
        raise NotImplementedError

    def delete_document(self, index_name: str, doc_id: str | int) -> None:
        raise NotImplementedError


class BaseIndexSearchBackend:
    """
    Abstract base search backend class for index management.
    """

    def initialize_indices(self, force_new_index: bool = False) -> None:
        raise NotImplementedError

    def rebuild_indices(
        self, batch_size: int = 500, extra_catchup_minutes: int = 5
    ) -> None:
        raise NotImplementedError

    def validate_indices(self) -> None:
        raise NotImplementedError

    def refresh_indices(self) -> None:
        """
        Make sure that all pending changes are applied to the indices.

        This is mostly used in tests.
        """
        raise NotImplementedError

    def delete_unused_indices(self) -> int:
        raise NotImplementedError


class BaseThreadSearchBackend:
    """
    Base class to perform thread search.
    """

    def get_thread_ids(
        self,
        context: str,
        group_ids: list[int],
        search_text: str,
        sort_criteria: t.Optional[list[dict[str, str]]] = None,
        commentable_ids: t.Optional[list[str]] = None,
        course_id: t.Optional[str] = None,
    ) -> list[str]:
        """
        Retrieve thread IDs based on search criteria.
        """
        raise NotImplementedError

    def get_suggested_text(self, search_text: str) -> t.Optional[str]:
        """
        Retrieve text suggestions for a given search query.

        :param search_text: Text to search for suggestions
        :return: Suggested text or None
        """
        raise NotImplementedError

    def get_thread_ids_with_corrected_text(
        self,
        context: str,
        group_ids: list[int],
        search_text: str,
        sort_criteria: t.Optional[list[dict[str, str]]] = None,
        commentable_ids: t.Optional[list[str]] = None,
        course_id: t.Optional[str] = None,
    ) -> list[str]:
        """
        The function is just used of mimicking the behaviour of the test cases.
        It's same as of the get_thread_ids but it could be used in the test cases for
        updating the returned values.
        """
        return self.get_thread_ids(
            context,
            group_ids,
            search_text,
            sort_criteria,
            commentable_ids,
            course_id,
        )


class BaseSearchBackend:
    """
    Abstract base search backend that exposes all search backend features.
    """

    DOCUMENT_SEARCH_CLASS: t.Type[BaseDocumentSearchBackend]
    INDEX_SEARCH_CLASS: t.Type[BaseIndexSearchBackend]
    THREAD_SEARCH_CLASS: t.Type[BaseThreadSearchBackend]



"""
Native Python Threads APIs.
"""

import logging
from typing import Any, Optional

from django.core.exceptions import ObjectDoesNotExist
from rest_framework.serializers import ValidationError

from forum.api.users import mark_thread_as_read
from forum.backend import get_backend
from forum.serializers.thread import ThreadSerializer
from forum.utils import ForumV2RequestError, get_int_value_from_collection, str_to_bool

log = logging.getLogger(__name__)


def _get_thread_data_from_request_data(data: dict[str, Any]) -> dict[str, Any]:
    """convert request data to a dict excluding empty data"""
    fields = [
        "title",
        "body",
        "course_id",
        "anonymous",
        "anonymous_to_peers",
        "closed",
        "commentable_id",
        "thread_type",
        "edit_reason_code",
        "close_reason_code",
        "endorsed",
        "pinned",
        "group_id",
        "context",
    ]
    result = {field: data.get(field) for field in fields if data.get(field) is not None}

    # Handle special cases
    if "user_id" in data:
        result["author_id"] = data["user_id"]
    if "editing_user_id" in data:
        result["editing_user_id"] = data["editing_user_id"]
    if "closing_user_id" in data:
        result["closed_by_id"] = data["closing_user_id"]

    return result


def get_thread_data(thread: dict[str, Any]) -> dict[str, Any]:
    """Prepare thread data for the api response."""
    _type = str(thread.get("_type", "")).lower()
    thread_data = {
        **thread,
        "id": str(thread.get("_id")),
        "type": "thread" if _type.lower() == "commentthread" else _type,
        "user_id": thread.get("author_id"),
        "username": str(thread.get("author_username")),
        "comments_count": thread["comment_count"],
    }
    return thread_data


def prepare_thread_api_response(
    thread: dict[str, Any],
    backend: Any,
    include_context: Optional[bool] = False,
    data_or_params: Optional[dict[str, Any]] = None,
    include_data_from_params: Optional[bool] = False,
    course_id: Optional[str] = None,
) -> dict[str, Any]:
    """Serialize thread data for the api response."""
    thread_data = get_thread_data(thread)

    context = {}
    if include_context:
        context = {
            "include_endorsed": True,
            "include_read_state": True,
        }
        if data_or_params:
            if user_id := data_or_params.get("user_id"):
                context["user_id"] = user_id

            if include_data_from_params:
                thread_data["resp_skip"] = get_int_value_from_collection(
                    data_or_params, "resp_skip", 0
                )
                thread_data["resp_limit"] = get_int_value_from_collection(
                    data_or_params, "resp_limit", 100
                )
                params = [
                    "recursive",
                    "with_responses",
                    "mark_as_read",
                    "reverse_order",
                    "merge_question_type_responses",
                ]
                for param in params:
                    if value := data_or_params.get(param):
                        context[param] = str_to_bool(value)
                if user_id and backend.get_user(user_id):
                    mark_thread_as_read(user_id, thread["_id"], course_id=course_id)

    serializer = ThreadSerializer(
        data=thread_data,
        context=context,
        backend=backend,
    )
    if not serializer.is_valid(raise_exception=True):
        log.error(f"validation error in thread API call: {serializer.errors}")
        raise ValidationError(serializer.errors)

    return serializer.data


def get_thread(
    thread_id: str,
    params: Optional[dict[str, Any]] = None,
    course_id: Optional[str] = None,
) -> dict[str, Any]:
    """
    Get the thread for the given thread_id.

    Parameters:
        thread_id: The ID of the thread.
        user_id: The ID of the user requesting the thread.
        resp_skip: Number of responses to skip.
        resp_limit: Maximum number of responses to return.
        recursive: Whether to include nested responses.
        with_responses: Whether to include responses.
        mark_as_read: Whether to mark the thread as read.
        reverse_order: Whether to reverse the order of responses.
        merge_question_type_responses: Whether to merge question type responses.
    Response:
        The details of the thread for the given thread_id.
    """
    backend = get_backend(course_id)()
    try:
        thread = backend.validate_object("CommentThread", thread_id)
    except ObjectDoesNotExist as exc:
        log.error("Forumv2RequestError for get thread request.")
        raise ForumV2RequestError(
            f"Thread does not exist with Id: {thread_id}"
        ) from exc

    try:
        return prepare_thread_api_response(
            thread,
            backend,
            True,
            params,
            True,
            course_id=course_id,
        )
    except ValidationError as error:
        log.error(f"Validation error in get_thread: {error}")
        raise ForumV2RequestError("Failed to prepare thread API response") from error


def delete_thread(thread_id: str, course_id: Optional[str] = None) -> dict[str, Any]:
    """
    Delete the thread for the given thread_id.

    Parameters:
        thread_id: The ID of the thread to be deleted.
    Response:
        The details of the thread that is deleted.
    """
    backend = get_backend(course_id)()
    try:
        thread = backend.validate_object("CommentThread", thread_id)
    except ObjectDoesNotExist as exc:
        log.error("Forumv2RequestError for delete thread request.")
        raise ForumV2RequestError(
            f"Thread does not exist with Id: {thread_id}"
        ) from exc

    backend.delete_comments_of_a_thread(thread_id)
    thread = backend.validate_object("CommentThread", thread_id)

    try:
        serialized_data = prepare_thread_api_response(thread, backend)
    except ValidationError as error:
        log.error(f"Validation error in get_thread: {error}")
        raise ForumV2RequestError("Failed to prepare thread API response") from error

    backend.delete_subscriptions_of_a_thread(thread_id)
    result = backend.delete_thread(thread_id)
    if result and not (thread["anonymous"] or thread["anonymous_to_peers"]):
        backend.update_stats_for_course(
            thread["author_id"], thread["course_id"], threads=-1
        )

    return serialized_data


def update_thread(
    thread_id: str,
    title: Optional[str] = None,
    body: Optional[str] = None,
    course_id: Optional[str] = None,
    anonymous: Optional[bool] = None,
    anonymous_to_peers: Optional[bool] = None,
    closed: Optional[bool] = None,
    commentable_id: Optional[str] = None,
    user_id: Optional[str] = None,
    editing_user_id: Optional[str] = None,
    pinned: Optional[bool] = None,
    thread_type: Optional[str] = None,
    edit_reason_code: Optional[str] = None,
    close_reason_code: Optional[str] = None,
    closing_user_id: Optional[str] = None,
    endorsed: Optional[bool] = None,
    course_key: Optional[str] = None,
) -> dict[str, Any]:
    """
    Update the thread for the given thread_id.

    Parameters:
        thread_id: The ID of the thread to be updated.
        data: The data to be updated.
    Response:
        The details of the thread that is updated.
    """
    backend = get_backend(course_key)()
    try:
        thread = backend.validate_object("CommentThread", thread_id)
    except ObjectDoesNotExist as exc:
        log.error("Forumv2RequestError for update thread request.")
        raise ForumV2RequestError(
            f"Thread does not exist with Id: {thread_id}"
        ) from exc

    data = {
        "title": title,
        "body": body,
        "course_id": course_id,
        "anonymous": anonymous,
        "anonymous_to_peers": anonymous_to_peers,
        "closed": closed,
        "commentable_id": commentable_id,
        "user_id": user_id,
        "editing_user_id": editing_user_id,
        "pinned": pinned,
        "thread_type": thread_type,
        "edit_reason_code": edit_reason_code,
        "close_reason_code": close_reason_code,
        "closing_user_id": closing_user_id,
        "endorsed": endorsed,
    }
    update_thread_data: dict[str, Any] = _get_thread_data_from_request_data(data)

    if "body" in update_thread_data:
        update_thread_data["original_body"] = thread.get("body")

    if update_thread_data.get("closed"):
        missing_fields = {"close_reason_code", "closed_by_id"} - set(
            update_thread_data.keys()
        )
        if missing_fields:
            raise ForumV2RequestError(
                f"Missing required fields: {', '.join(missing_fields)}"
            )
    backend.update_thread(thread_id, **update_thread_data)
    thread = backend.get_thread(thread_id)

    try:
        return prepare_thread_api_response(
            thread,
            backend,
            True,
            data,
        )
    except ValidationError as error:
        log.error(f"Validation error in get_thread: {error}")
        raise ForumV2RequestError("Failed to prepare thread API response") from error


def create_thread(
    title: str,
    body: str,
    course_id: str,
    user_id: str,
    anonymous: bool = False,
    anonymous_to_peers: bool = False,
    commentable_id: str = "course",
    thread_type: str = "discussion",
    group_id: Optional[int] = None,
    context: str = "course",
) -> dict[str, Any]:
    """
    Create a new thread.

    Parameters:
        title: The title of the thread.
        body: The body of the thread.
        course_id: The ID of the course.
        anonymous: Whether the thread is anonymous.
        anonymous_to_peers: Whether the thread is anonymous to peers.
        closed: Whether the thread is closed.
        commentable_id: The ID of the commentable.
        user_id: The ID of the user.
        group_id: The ID of the group.
    Response:
        The details of the thread that is created.
    """
    backend = get_backend(course_id)()
    data = {
        "title": title,
        "body": body,
        "course_id": course_id,
        "user_id": user_id,
        "anonymous": anonymous,
        "anonymous_to_peers": anonymous_to_peers,
        "commentable_id": commentable_id,
        "thread_type": thread_type,
        "group_id": group_id,
        "context": context,
    }
    thread_data: dict[str, Any] = _get_thread_data_from_request_data(data)

    thread_id = backend.create_thread(thread_data)
    thread = backend.get_thread(thread_id)
    if not thread:
        raise ForumV2RequestError(f"Failed to create thread with data: {data}")

    if not (anonymous or anonymous_to_peers):
        backend.update_stats_for_course(
            thread["author_id"], thread["course_id"], threads=1
        )

    try:
        return prepare_thread_api_response(
            thread,
            backend,
            True,
            data,
        )
    except ValidationError as error:
        log.error(f"Validation error in get_thread: {error}")
        raise ForumV2RequestError("Failed to prepare thread API response") from error


def get_user_threads(
    course_id: str,
    author_id: Optional[str] = None,
    thread_type: Optional[str] = None,
    flagged: Optional[bool] = None,
    unread: Optional[bool] = None,
    unanswered: Optional[bool] = None,
    unresponded: Optional[bool] = None,
    count_flagged: Optional[bool] = None,
    sort_key: Optional[str] = None,
    page: Optional[str] = None,
    per_page: Optional[str] = None,
    request_id: Optional[str] = None,
    commentable_ids: Optional[str] = None,
    user_id: Optional[str] = None,
    group_id: Optional[int] = None,
    group_ids: Optional[int] = None,
    **kwargs: Any,
) -> dict[str, Any]:
    """
    Get the threads for the given thread_ids.
    """
    backend = get_backend(course_id)()
    params = {
        "course_id": course_id,
        "author_id": author_id,
        "thread_type": thread_type,
        "flagged": flagged,
        "unread": unread,
        "unanswered": unanswered,
        "unresponded": unresponded,
        "count_flagged": count_flagged,
        "sort_key": sort_key,
        "page": int(page) if page else None,
        "per_page": int(per_page) if per_page else None,
        "request_id": request_id,
        "commentable_ids": commentable_ids,
        "user_id": user_id,
        "group_id": group_id,
        "group_ids": group_ids,
    }
    params = {k: v for k, v in params.items() if v is not None}
    backend.validate_params(params)

    thread_filter = backend.get_user_thread_filter(course_id)
    filtered_threads = backend.get_filtered_threads(thread_filter)
    thread_ids = [thread["_id"] for thread in filtered_threads]
    threads = backend.get_threads(params, user_id or "", ThreadSerializer, thread_ids)

    return threads


def get_course_id_by_thread(thread_id: str) -> str | None:
    """
    Return course_id for the matching thread.
    It searches for thread_id both in mongodb and mysql.
    """
    #  pylint: disable=C0415
    from forum.backends.mongodb.api import MongoBackend
    from forum.backends.mysql.api import MySQLBackend

    return (
        MongoBackend.get_course_id_by_thread_id(thread_id)
        or MySQLBackend.get_course_id_by_thread_id(thread_id)
        or None
    )



"""Subscriptions API Views."""

import logging
from typing import Any

from rest_framework import status
from rest_framework.exceptions import ParseError
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from forum.api import get_user
from forum.api.users import (
    create_user,
    get_user_active_threads,
    get_user_course_stats,
    mark_thread_as_read,
    retire_user,
    update_user,
    update_username,
    update_users_in_course,
)
from forum.utils import ForumV2RequestError, get_group_ids_from_params, str_to_bool

log = logging.getLogger(__name__)


class UserAPIView(APIView):
    """Users API View."""

    permission_classes = (AllowAny,)

    def get(self, request: Request, user_id: str) -> Response:
        """Get user data."""
        params = request.GET.dict()
        course_id = params.get("course_id", "")
        group_ids = get_group_ids_from_params(params)
        complete = str_to_bool(params.get("complete", False))

        try:
            user_data: dict[str, Any] = get_user(
                user_id, group_ids, course_id, complete
            )
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_404_NOT_FOUND)

        return Response(user_data, status=status.HTTP_200_OK)

    def put(self, request: Request, user_id: str) -> Response:
        """Update user data."""
        try:
            params = request.data
            username = params.get("username")
            default_sort_key = params.get("default_sort_key")
            course_id = params.get("course_id")
            group_ids = params.get("group_ids")
            complete = params.get("complete")
            updated_user = update_user(
                user_id,
                username,
                default_sort_key,
                course_id,
                group_ids,
                complete,
            )
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)

        return Response(updated_user, status=status.HTTP_200_OK)


class UserCreateAPIView(APIView):
    """Create users api."""

    permission_classes = (AllowAny,)

    def post(self, request: Request) -> Response:
        """Create user."""
        params = request.data
        for key in params:
            if key not in ["id", "username"]:
                return Response(
                    {"error": f"Invalid parameter: {key}"},
                    status=status.HTTP_400_BAD_REQUEST,
                )
        try:
            data: dict[str, Any] = {
                "user_id": params.get("id"),
                "username": params.get("username"),
                "default_sort_key": params.get("default_sort_key", "date"),
                "course_id": params.get("course_id"),
                "group_ids": params.get("group_ids"),
                "complete": params.get("complete"),
            }
            user_data = create_user(**data)
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)
        return Response(user_data, status=status.HTTP_200_OK)


class UserEditAPIView(APIView):
    """Edit user api."""

    def post(self, request: Request, user_id: str) -> Response:
        """Edit user."""
        error_500_response = Response(
            {"message": "Missing new_username param."},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )
        try:
            params = request.data
        except ParseError:
            return error_500_response
        new_username = params.get("new_username")
        if not new_username:
            return error_500_response
        try:
            update_username(user_id, new_username)
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)
        return Response(status=status.HTTP_200_OK)


class UserRetireAPIView(APIView):
    """Retire user api."""

    permission_classes = (AllowAny,)

    def post(self, request: Request, user_id: str) -> Response:
        """Retire user."""
        error_500_response = Response(
            {"message": "Missing retired_username param."},
            status=status.HTTP_500_INTERNAL_SERVER_ERROR,
        )
        try:
            params = request.data
        except ParseError:
            return error_500_response
        retired_username = params.get("retired_username")
        if not retired_username:
            return error_500_response
        try:
            retire_user(user_id, retired_username)
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)
        return Response(status=status.HTTP_200_OK)


class UserReadAPIView(APIView):
    """User read api."""

    permission_classes = (AllowAny,)

    def post(self, request: Request, user_id: str) -> Response:
        """User read."""
        params = request.data
        data = {
            "source_id": params.get("source_id", ""),
            "complete": params.get("complete"),
            "course_id": params.get("course_id"),
            "group_ids": params.get("group_ids"),
        }
        try:
            serialized_data = mark_thread_as_read(user_id, **data)
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)
        return Response(serialized_data, status=status.HTTP_200_OK)


class UserActiveThreadsAPIView(APIView):
    """User active threads api."""

    permission_classes = (AllowAny,)

    def get(self, request: Request, user_id: str) -> Response:
        """User active threads."""
        params: dict[str, Any] = request.GET.dict()
        course_id = params.pop("course_id", None)

        if page := params.get("page"):
            params["page"] = int(page)
        if per_page := params.get("per_page"):
            params["per_page"] = int(per_page)
        if flagged := params.get("flagged"):
            params["flagged"] = str_to_bool(flagged)
        if unread := params.get("unread"):
            params["unread"] = str_to_bool(unread)
        if unanswered := params.get("unanswered"):
            params["unanswered"] = str_to_bool(unanswered)
        if unresponded := params.get("unresponded"):
            params["unresponded"] = str_to_bool(unresponded)
        if count_flagged := params.get("count_flagged"):
            params["count_flagged"] = str_to_bool(count_flagged)
        if group_id := params.get("group_id"):
            params["group_id"] = int(group_id)
        try:
            serialized_data = get_user_active_threads(user_id, course_id, **params)
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)
        return Response(serialized_data)


class UserCourseStatsAPIView(APIView):
    """User Course stats API."""

    permission_classes = (AllowAny,)

    def get(self, request: Request, course_id: str) -> Response:
        """Get user course stats."""
        params: dict[str, Any] = request.GET.dict()
        if page := params.get("page"):
            params["page"] = int(page)
        if per_page := params.get("per_page"):
            params["per_page"] = int(per_page)
        if with_timestamps := params.get("with_timestamps"):
            params["with_timestamps"] = str_to_bool(with_timestamps)

        response = get_user_course_stats(
            course_id,
            **params,
        )
        return Response(response, status=status.HTTP_200_OK)

    def post(self, request: Request, course_id: str) -> Response:
        """Update user stats for a course."""
        updated_users = update_users_in_course(course_id)
        return Response(updated_users, status=status.HTTP_200_OK)



"""Migration command for courses from mongodb to mysql."""

from typing import Any

from django.core.management.base import BaseCommand
from django.core.management.base import CommandParser

from forum.migration_helpers import (
    enable_mysql_backend_for_course,
    get_all_course_ids,
    migrate_content,
    migrate_read_states,
    migrate_users,
)
from forum.mongo import get_database


class Command(BaseCommand):
    """Migration command for courses from mongodb to mysql."""

    help = "Migrate data from MongoDB to MySQL"

    def add_arguments(self, parser: CommandParser) -> None:
        """Add arguments to the command."""
        parser.add_argument(
            "-T",
            "--no-toggle",
            action="store_true",
            help="Skip course waffle flag creation",
        )
        parser.add_argument(
            "courses", nargs="+", type=str, help="List of course IDs or `all`"
        )

    def handle(self, *args: str, **options: dict[str, Any]) -> None:
        """Handle the command."""
        db = get_database()

        create_waffle_flags = not options["no_toggle"]
        course_ids = list(options["courses"])
        if "all" in course_ids:
            course_ids = get_all_course_ids(db)

        for course_id in course_ids:
            self.stdout.write(f"Migrating data for course: {course_id}")
            migrate_users(db, course_id)
            migrate_content(db, course_id)
            migrate_read_states(db, course_id)
            if create_waffle_flags:
                enable_mysql_backend_for_course(course_id)
                self.stdout.write(
                    f"Enabled mysql backend waffle flag for course {course_id}."
                )

        self.stdout.write(self.style.SUCCESS("Data migration completed successfully"))



"""
Elasticsearch client utilities.
"""

import logging
import re
from datetime import datetime, timedelta
from typing import Any, Iterator, Optional

from django.conf import settings
from elasticsearch import Elasticsearch, exceptions, helpers

from forum.backends.mongodb import MODEL_INDICES as mongo_model_indices
from forum.backends.mongodb import BaseContents
from forum.backends.mongodb.threads import CommentThread
from forum.backends.mysql import MODEL_INDICES as mysql_model_indices
from forum.constants import FORUM_MAX_DEEP_SEARCH_COMMENT_COUNT
from forum.models import Content
from forum.search import base

__all__ = ["ElasticsearchBackend"]

log = logging.getLogger(__name__)


class ElasticsearchClientMixin:
    """
    Provide a Elasticsearch client API based on a singleton.
    """

    ELASTIC_SEARCH_INSTANCE: Optional[Elasticsearch] = None

    @property
    def client(self) -> Elasticsearch:
        """
        Elasticsearch client singleton.
        """
        if self.ELASTIC_SEARCH_INSTANCE is None:
            self.ELASTIC_SEARCH_INSTANCE = Elasticsearch(
                settings.FORUM_ELASTIC_SEARCH_CONFIG
            )
        return self.ELASTIC_SEARCH_INSTANCE


class ElasticsearchModelMixin:
    """
    Provides convenient model and index name properties.
    """

    @property
    def models(self) -> tuple[type[BaseContents], ...]:
        return mongo_model_indices

    @property
    def mysql_models(self) -> tuple[type[Content], ...]:
        return mysql_model_indices

    @property
    def index_names(self) -> list[str]:
        """
        Retrieve the base index names for the configured models.

        Returns:
            list[str]: List of base index names.
        """
        return [model.index_name for model in self.models]

    def get_mysql_model_from_index_name(self, index_name: str) -> type[Content]:
        for model in self.mysql_models:
            if index_name.startswith(model.index_name):
                return model
        raise Exception("Invalid model name")


class ElasticsearchDocumentBackend(
    base.BaseDocumentSearchBackend, ElasticsearchClientMixin
):
    """
    Elasticsearch-based backend for document management.
    """

    def update_document(
        self, index_name: str, doc_id: str | int, update_data: dict[str, Any]
    ) -> None:
        """
        Update a single document in the specified index.

        Args:
            index_name (str): The name of the index containing the document.
            doc_id (str): The ID of the document to update.
            update_data (dict): The data to update in the document.
        """
        try:
            self.client.update(index=index_name, id=doc_id, body={"doc": update_data})
            log.info(f"Document {doc_id} in index {index_name} updated successfully.")
        except exceptions.NotFoundError:
            log.error(f"Document {doc_id} not found in index {index_name}.")
        except exceptions.RequestError as e:
            log.error(f"Error updating document {doc_id} in index {index_name}: {e}")

    def delete_document(self, index_name: str, doc_id: str | int) -> None:
        """
        Delete a single document from the specified index.

        Args:
            index_name (str): The name of the index containing the document.
            doc_id (str): The ID of the document to delete.
        """
        try:
            self.client.delete(index=index_name, id=doc_id)
            log.info(f"Document {doc_id} in index {index_name} deleted successfully.")
        except exceptions.NotFoundError:
            log.error(f"Document {doc_id} not found in index {index_name}.")
        except exceptions.RequestError as e:
            log.error(f"Error deleting document {doc_id} from index {index_name}: {e}")

    def index_document(
        self, index_name: str, doc_id: str | int, document: dict[str, Any]
    ) -> None:
        """
        Index a single document in the specified Elasticsearch index.

        Args:
            index_name (str): The name of the index to add the document to.
            doc_id (str): The ID of the document.
            document (dict): The document to be indexed.
        """
        try:
            self.client.index(index=index_name, id=doc_id, body=document)
            log.info(f"Document {doc_id} indexed in {index_name}")
        except exceptions.RequestError as e:
            log.error(f"Error indexing document {doc_id} in {index_name}: {e}")


class ElasticsearchIndexBackend(
    base.BaseIndexSearchBackend, ElasticsearchModelMixin, ElasticsearchClientMixin
):
    """
    Search forum threads.
    """

    MAPPINGS: dict[str, dict[str, Any]] = {
        "comments": {
            "dynamic": "false",
            "properties": {
                "body": {
                    "type": "text",
                    "store": True,
                    "term_vector": "with_positions_offsets",
                },
                "course_id": {"type": "keyword"},
                "comment_thread_id": {"type": "keyword"},
                "commentable_id": {"type": "keyword"},
                "group_id": {"type": "keyword"},
                "context": {"type": "keyword"},
                "created_at": {"type": "date"},
                "updated_at": {"type": "date"},
                "title": {"type": "keyword"},
            },
        },
        "comment_threads": {
            "dynamic": "false",
            "properties": {
                "title": {
                    "type": "text",
                    "boost": 5.0,
                    "store": True,
                    "term_vector": "with_positions_offsets",
                },
                "body": {
                    "type": "text",
                    "store": True,
                    "term_vector": "with_positions_offsets",
                },
                "created_at": {"type": "date"},
                "updated_at": {"type": "date"},
                "last_activity_at": {"type": "date"},
                "comment_count": {"type": "integer"},
                "votes_point": {"type": "integer"},
                "context": {"type": "keyword"},
                "course_id": {"type": "keyword"},
                "commentable_id": {"type": "keyword"},
                "author_id": {"type": "keyword"},
                "group_id": {"type": "integer"},
                "id": {"type": "keyword"},
                "thread_id": {"type": "keyword"},
            },
        },
    }

    def rebuild_indices(
        self, batch_size: int = 500, extra_catchup_minutes: int = 5
    ) -> None:
        """
        Rebuild the indices by creating new indices, importing data, and managing aliases.

        Args:
            batch_size (int): Number of documents to process in each batch.
            extra_catchup_minutes (int): Extra minutes to adjust the start time for catch-up.
        """
        initial_start_time = datetime.now()

        # Create new indices and switch aliases
        index_names = self.create_indices()
        for index_name in index_names:
            current_batch = 1
            mongo_model = self.get_index_model_rel(index_name)
            for response in self._import_to_es_from_mongo(
                mongo_model, index_name, batch_size
            ):
                self.batch_import_post_process(response, current_batch)
                current_batch += 1

        for index_name in index_names:
            current_batch = 1
            mysql_model = self.get_mysql_model_from_index_name(index_name)
            for response in self._import_to_es_from_mysql(
                mysql_model, index_name, batch_size
            ):
                self.batch_import_post_process(response, current_batch)
                current_batch += 1

        adjusted_start_time = initial_start_time - timedelta(
            minutes=extra_catchup_minutes
        )
        self.catchup_indices(index_names, adjusted_start_time, batch_size)

        # Update aliases to point to new indices
        for index_name in index_names:
            model = self.get_index_model_rel(index_name)
            self.move_alias(model.index_name, index_name, force_delete=True)

        self.delete_unused_indices()

        log.info("Rebuild indices complete.")

    def get_index_model_rel(self, index_name: str) -> BaseContents:
        """
        Retrieve the model corresponding to the given index name.

        Args:
            index_name (str): Name of the index.

        Returns:
            model: The model associated with the index name.
        """
        for model in self.models:
            if index_name.startswith(model.index_name):
                return model()
        raise ValueError("Invalid index name")

    def catchup_indices(
        self, index_names: list[str], start_time: datetime, batch_size: int = 100
    ) -> None:
        """
        Catch up the indices by importing documents updated after the specified start time.

        Args:
            index_names (list[str]): List of index names to catch up.
            start_time (datetime): The starting time for catching up.
            batch_size (int): Number of documents to process in each batch.
        """
        for index_name in index_names:
            current_batch = 1
            mongo_model = self.get_index_model_rel(index_name)
            mongo_query: dict[str, Any] = {"updated_at": {"$gte": start_time}}
            for response in self._import_to_es_from_mongo(
                mongo_model, index_name, batch_size, mongo_query
            ):
                self.batch_import_post_process(response, current_batch)
                current_batch += 1

        for index_name in index_names:
            current_batch = 1
            mysql_model = self.get_mysql_model_from_index_name(index_name)
            mysql_query: dict[str, Any] = {"updated_at__gte": start_time}
            for response in self._import_to_es_from_mysql(
                mysql_model, index_name, batch_size, mysql_query
            ):
                self.batch_import_post_process(response, current_batch)
                current_batch += 1
        log.info(f"Catch up from {start_time} complete.")

    def create_indices(self) -> list[str]:
        """
        Create new indices for each model with the current timestamp in the index name.

        Returns:
            list[str]: List of newly created index names.
        """
        index_names = []
        time_now = datetime.now().strftime("%Y%m%d%H%M%S")

        for model in self.models:
            index_name = f"{model.index_name}_{time_now}"
            index_names.append(index_name)
            self.client.indices.create(
                index=index_name,
                body={"mappings": self.MAPPINGS[model.index_name]},
            )
        log.info(f"New indices {index_names} are created.")
        return index_names

    def delete_index(self, name: str) -> None:
        """
        Delete the specified index.

        Args:
            name (str): The name of the index to delete.
        """
        self.client.indices.delete(index=name)
        log.info(f"Deleted index: {name}.")

    def delete_unused_indices(self) -> int:
        """
        Delete all Elasticsearch indices that are not the latest for each model type.

        This method fetches all indices matching the patterns from the models, determines the latest index
        for each model type, and deletes all other indices.

        Returns:
            int: The number of indices deleted.
        """
        # Fetch all indices related to models
        all_indices = [
            index
            for pattern in self.index_names
            for index in self.client.indices.get(f"{pattern}*")
        ]

        # Determine the latest indices
        latest_indices: dict[str, Any] = {}
        for index_name in all_indices:
            base_name = self.get_base_index_name(index_name)
            match = re.search(r"\d{14}", index_name)
            if match:
                timestamp = datetime.strptime(match.group(), "%Y%m%d%H%M%S")
                if (
                    base_name not in latest_indices
                    or timestamp > latest_indices[base_name][1]
                ):
                    latest_indices[base_name] = (index_name, timestamp)

        # Delete all indices except the latest ones
        indices_to_delete = set(all_indices) - {
            name for name, _ in latest_indices.values()
        }
        if indices_to_delete:
            self.client.indices.delete(index=",".join(indices_to_delete))
            log.info(f"Deleted unused indices: {indices_to_delete}")

        return len(indices_to_delete)

    def batch_import_post_process(
        self, response: tuple[int, Any], batch_number: int
    ) -> None:
        """
        Process the response from a batch import operation.

        Args:
            response (tuple[int, list[dict]]): tuple containing the count of successful imports and a list of errors.
            batch_number (int): The current batch number being processed.
        """
        success_count, errors = response
        for item in errors:
            if "error" in item["index"]:
                log.error(f"Error indexing. Response was: {response}")
        log.info(
            f"Imported {success_count} documents to the batch {batch_number} into the index"
        )

    def move_alias(
        self, alias_name: str, index_name: str, force_delete: bool = False
    ) -> None:
        """
        Move an alias to point to a new index, optionally deleting an existing index with the same name.

        Args:
            alias_name (str): The alias name to move.
            index_name (str): The new index name to point the alias to.
            force_delete (bool): If True, deletes any existing index with the alias name before moving the alias.

        Raises:
            ValueError: If the alias name matches the index name or if the index doesn't exist.
        """
        if alias_name == index_name:
            raise ValueError(
                f"Can't point alias [{alias_name}] to an index of the same name."
            )
        if not self.exists_index(index_name):
            raise ValueError(f"Can't point alias to non-existent index [{index_name}].")

        if self.exists_index(alias_name) and not self.exists_alias(alias_name):
            if force_delete:
                self.delete_index(alias_name)
            else:
                raise ValueError(
                    f"Can't create alias [{alias_name}] because there is already an "
                    f"auto-generated index of the same name. Try again with force_delete=True to first "
                    f"delete this pre-existing index."
                )

        actions = [{"add": {"index": index_name, "alias": alias_name}}]

        try:
            response = self.client.indices.get_alias(name=alias_name)
            if response:
                actions.insert(
                    0,
                    {
                        "remove": {
                            "index": ",".join(response.keys()),
                            "alias": alias_name,
                        }
                    },
                )
        except exceptions.NotFoundError as e:
            log.warning(f"Alias not found for {alias_name}: {e}")

        self.client.indices.update_aliases(body={"actions": actions})
        log.info(f"Alias [{alias_name}] now points to index [{index_name}].")

    def refresh_indices(self) -> None:
        """
        Refresh the indices associated with the current active aliases.

        Raises:
            ValueError: If no indices are available to refresh.
        """
        active_index_names = self.get_active_index_names()
        if active_index_names:
            self.client.indices.refresh(index=active_index_names)
        else:
            raise ValueError("No indices to refresh")

    def initialize_indices(self, force_new_index: bool = False) -> None:
        """
        Initialize the indices by creating new indices or reusing existing ones based on aliases.

        Args:
            force_new_index (bool): If True, forces the creation of new indices regardless of existing ones.
        """
        if force_new_index or not self.exists_aliases(self.index_names):
            index_names = self.create_indices()
            for index_name in index_names:
                model = self.get_index_model_rel(index_name)
                self.move_alias(model.index_name, index_name, force_delete=True)
        else:
            log.info("Skipping initialization. Indices already exist.")

    def validate_indices(self) -> None:
        """
        Validate that the indices have the correct mappings and properties.

        Raises:
            ValueError: If indices do not exist, or mappings/properties are missing or incorrect.
        """
        actual_mappings = self.client.indices.get_mapping(index=self.index_names)

        if not actual_mappings:
            raise ValueError("Indices do not exist!")

        for index_name in actual_mappings:
            model = self.get_index_model_rel(index_name)
            expected_mapping = self.MAPPINGS[model.index_name]
            actual_mapping = actual_mappings[index_name]["mappings"]

            if not expected_mapping.items() <= actual_mapping.items():
                raise ValueError(f"Mapping mismatch for index {index_name}!")

        log.info("Index validation complete.")

    def exists_index(self, name: str) -> bool:
        """
        Check if an index exists.

        Args:
            name (str): The name of the index to check.

        Returns:
            bool: True if the index exists, False otherwise.
        """
        return self.client.indices.exists(index=name)

    def exists_alias(self, name: str) -> bool:
        """
        Check if an alias exists.

        Args:
            name (str): The name of the alias to check.

        Returns:
            bool: True if the alias exists, False otherwise.
        """
        return self.client.indices.exists_alias(name=name)

    def exists_aliases(self, names: list[str]) -> bool:
        """
        Check if all aliases in the provided list exist.

        Args:
            names (list[str]): List of alias names to check.

        Returns:
            bool: True if all aliases exist, False otherwise.
        """
        return all(self.exists_alias(name) for name in names)

    def _import_to_es_from_mongo(
        self,
        model: BaseContents,
        index_name: str,
        batch_size: int = 500,
        query: dict[str, Any] | None = None,
    ) -> Iterator[tuple[int, Any]]:
        """
        Import documents from the database into Elasticsearch.

        Args:
            model (BaseContents): The model representing the documents.
            index_name (str): The name of the index to import into.
            batch_size (int): Number of documents to import in each batch.
            query (dict[str, Any], optional): Query to filter documents for import.

        Yields:
            Iterator[tuple[int, Any]]: Number of successful imports and any errors.
        """
        cursor = model.find(query or {}).batch_size(batch_size)
        actions = []
        for doc in cursor:
            action = {
                "_index": index_name,
                "_id": str(doc.get("_id")),
                "_source": model.doc_to_hash(doc),
            }
            actions.append(action)
            if len(actions) >= batch_size:
                yield helpers.bulk(self.client, actions)
                actions = []
        if actions:
            yield helpers.bulk(self.client, actions)

    def _import_to_es_from_mysql(
        self,
        model: Any,
        index_name: str,
        batch_size: int = 500,
        query: dict[str, Any] | None = None,
    ) -> Iterator[tuple[int, Any]]:
        """
        Import documents from the MySQL database into Elasticsearch.

        Args:
            model (Content): The model representing the documents.
            index_name (str): The name of the index to import into.
            batch_size (int): Number of documents to import in each batch.
            query (dict[str, Any], optional): Query to filter documents for import.

        Yields:
            Iterator[tuple[int, Any]]: Number of successful imports and any errors.
        """
        queryset = model.objects.all()
        if query:
            queryset = queryset.filter(**query)

        actions = []
        for doc in queryset.iterator(chunk_size=batch_size):
            action = {
                "_index": index_name,
                "_id": str(doc.id),
                "_source": model.doc_to_hash(doc),
            }
            actions.append(action)
            if len(actions) >= batch_size:
                yield helpers.bulk(self.client, actions)
                actions = []
        if actions:
            yield helpers.bulk(self.client, actions)

    def get_active_index_names(self) -> str:
        """
        Retrieve the names of the indices currently associated with the aliases.

        Returns:
            str: Comma-separated list of active index names.
        """
        return ",".join(
            [
                list(self.client.indices.get_alias(name=index_name).keys())[0]
                for index_name in self.index_names
                if self.exists_alias(index_name)
            ]
        )

    def get_base_index_name(self, index_name: str) -> str:
        """
        Extract the base name of an index by removing the timestamp.

        Args:
            index_name (str): The full index name.

        Returns:
            str: The base name of the index.
        """
        return re.sub(r"_\d{14}$", "", index_name)


class ElasticsearchThreadSearchBackend(
    base.BaseThreadSearchBackend, ElasticsearchClientMixin, ElasticsearchModelMixin
):
    """
    Base class to perform thread search.
    """

    def get_suggested_text(self, search_text: str) -> Optional[str]:
        """
        Retrieve text suggestions for a given search query.

        :param search_text: Text to search for suggestions
        :return: Suggested text or None
        """
        suggestion_fields = ["body", "title"]
        suggest_body: dict[str, Any] = {
            "suggest": {
                f"{field}_suggestions": {
                    "text": search_text,
                    "phrase": {"field": field},
                }
                for field in suggestion_fields
            }
        }
        response: dict[str, Any] = self.client.search(
            index=self.index_names, body=suggest_body
        )
        return self._extract_suggestion(
            response, [f"{field}_suggestions" for field in suggestion_fields]
        )

    @staticmethod
    def _extract_suggestion(
        response: dict[str, Any], suggestion_types: list[str]
    ) -> Optional[str]:
        """
        Extract suggestions from the Elasticsearch response.
        """
        for suggestion_type in suggestion_types:
            suggestions: list[dict[str, Any]] = response.get("suggest", {}).get(
                suggestion_type, []
            )
            options = suggestions and suggestions[0].get("options", [])
            if options:
                return options[0]["text"]
        return None

    def build_must_clause(
        self,
        search_text: str,
        commentable_ids: Optional[list[str]] = None,
        course_id: Optional[str] = None,
    ) -> list[dict[str, Any]]:
        """
        Build the 'must' clause for thread-specific Elasticsearch queries based on input parameters.
        """
        must: list[dict[str, Any]] = []
        commentable_ids = commentable_ids or []

        if len(commentable_ids) == 1:
            must.append({"term": {"commentable_id": commentable_ids[0]}})
        elif len(commentable_ids) > 1:
            must.append({"terms": {"commentable_id": commentable_ids}})
        if course_id:
            must.append({"term": {"course_id": course_id}})

        must.append(
            {
                "multi_match": {
                    "query": search_text,
                    "fields": ["title", "body"],
                    "operator": "AND",
                }
            }
        )

        return must

    def build_filter_clause(
        self,
        context: str,
        group_ids: Optional[list[int]] = None,
    ) -> list[dict[str, Any]]:
        """
        Build the 'filter' clause for thread-specific Elasticsearch queries based on context and group parameters.
        """
        filter_clause: list[dict[str, Any]] = []

        filter_clause.extend(
            [
                {"bool": {"must_not": {"exists": {"field": "context"}}}},
                {"term": {"context": context}},
            ]
        )

        if group_ids and len(group_ids) > 1:
            filter_clause.extend(
                [
                    {"bool": {"must_not": {"exists": {"field": "group_id"}}}},
                    {"terms": {"group_id": group_ids}},
                ]
            )

        elif group_ids and len(group_ids) == 1:
            filter_clause.extend(
                [
                    {"bool": {"must_not": {"exists": {"field": "group_id"}}}},
                    {"term": {"group_id": group_ids[0]}},
                ]
            )

        return filter_clause

    def get_thread_ids(
        self,
        context: str,
        group_ids: list[int],
        search_text: str,
        sort_criteria: Optional[list[dict[str, str]]] = None,
        commentable_ids: Optional[list[str]] = None,
        course_id: Optional[str] = None,
    ) -> list[str]:
        """
        Retrieve thread IDs based on search criteria.
        """
        must_clause: list[dict[str, Any]] = self.build_must_clause(
            search_text, commentable_ids, course_id
        )
        filter_clause: list[dict[str, Any]] = self.build_filter_clause(
            context, group_ids
        )
        search_body = {
            "size": FORUM_MAX_DEEP_SEARCH_COMMENT_COUNT,
            "sort": sort_criteria or [{"updated_at": "desc"}],
            "query": {
                "bool": {"must": must_clause or [], "should": filter_clause or []}
            },
        }
        response = self.client.search(index=self.index_names, body=search_body)
        if response:
            return list(
                {
                    (
                        hit["_id"]
                        if CommentThread.index_name in hit.get("_index")
                        else hit.get("_source", {}).get("comment_thread_id")
                    )
                    for hit in response.get("hits", {}).get("hits")
                }
            )

        return []


class ElasticsearchBackend(base.BaseSearchBackend):
    """
    Helper class for managing Forum indices.
    """

    DOCUMENT_SEARCH_CLASS = ElasticsearchDocumentBackend
    INDEX_SEARCH_CLASS = ElasticsearchIndexBackend
    THREAD_SEARCH_CLASS = ElasticsearchThreadSearchBackend



"""Forum Comments API Views."""

from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import ValidationError
from rest_framework.views import APIView

from forum.api import (
    create_parent_comment,
    create_child_comment,
    delete_comment,
    get_parent_comment,
    update_comment,
)
from forum.utils import ForumV2RequestError, str_to_bool


class CommentsAPIView(APIView):
    """
    API View to handle GET, POST, PUT, and DELETE requests for comments.
    """

    permission_classes = (AllowAny,)

    def get(self, request: Request, comment_id: str) -> Response:
        """
        Retrieves a parent comment.
        For chile comments, below API is called that return all child comments in children field
        url: http://localhost:8000/forum/api/v2/threads/66ab94950dead7001deb947a

        Parameters:
            request (Request): The incoming request.
            comment_id: The ID of the comment.
        Body:
            Empty.
        Response:
            The details of the comment for the given comment_id.
        """
        try:
            data = get_parent_comment(comment_id)
        except ForumV2RequestError:
            return Response(
                {"error": f"Comment does not exist with Id: {comment_id}"},
                status=status.HTTP_400_BAD_REQUEST,
            )
        return Response(data, status=status.HTTP_200_OK)

    def post(self, request: Request, comment_id: str) -> Response:
        """
        Creates a new child comment.
        For parent comment below API is called.
        url: http://localhost:8000/forum/api/v2/threads/66ab94950dead7001deb947a/comments

        Parameters:
            request (Request): The incoming request.
            comment_id: The ID of the parent comment for creating it's child comment.
        Body:
            body: The content of the comment.
            course_id: The Id of the respective course.
            user_id: The requesting user id.
            anonymous: anonymous flag(True or False).
            anonymous_to_peers: anonymous to peers flag(True or False).
        Response:
            The details of the comment that is created.
        """
        try:
            request_data = request.data
            comment = create_child_comment(
                comment_id,
                request_data["body"],
                request_data["user_id"],
                request_data["course_id"],
                str_to_bool(request_data.get("anonymous", False)),
                str_to_bool(request_data.get("anonymous_to_peers", False)),
            )
        except ForumV2RequestError:
            return Response(
                {"error": f"Comment does not exist with Id: {comment_id}"},
                status=status.HTTP_400_BAD_REQUEST,
            )
        except ValidationError as e:
            return Response(
                {"error": e.detail},
                status=status.HTTP_400_BAD_REQUEST,
            )
        return Response(comment, status=status.HTTP_200_OK)

    def put(self, request: Request, comment_id: str) -> Response:
        """
        Updates an existing child/parent comment.

        Parameters:
            request (Request): The incoming request.
            comment_id: The ID of the comment to be edited.
        Body:
            fields to be updated.
        Response:
            The details of the comment that is updated.
        """
        try:
            request_data = request.data
            if anonymous := request_data.get("anonymous"):
                anonymous = str_to_bool(anonymous)
            if anonymous_to_peers := request_data.get("anonymous_to_peers"):
                anonymous_to_peers = str_to_bool(anonymous_to_peers)
            if endorsed := request_data.get("endorsed"):
                endorsed = str_to_bool(endorsed)
            if closed := request_data.get("closed"):
                closed = str_to_bool(closed)
            comment = update_comment(
                comment_id,
                request_data.get("body"),
                request_data.get("course_id"),
                request_data.get("user_id"),
                anonymous,
                anonymous_to_peers,
                endorsed,
                closed,
                request_data.get("editing_user_id"),
                request_data.get("edit_reason_code"),
                request_data.get("endorsement_user_id"),
            )
        except ForumV2RequestError:
            return Response(
                {"error": f"Comment does not exist with Id: {comment_id}"},
                status=status.HTTP_400_BAD_REQUEST,
            )
        except ValidationError as e:
            return Response(
                {"error": e.detail},
                status=status.HTTP_400_BAD_REQUEST,
            )
        return Response(comment, status=status.HTTP_200_OK)

    def delete(self, request: Request, comment_id: str) -> Response:
        """
        Deletes a comment.

        Parameters:
            request (Request): The incoming request.
            comment_id: The ID of the comment to be deleted.
        Body:
            Empty.
        Response:
            The details of the comment that is deleted.
        """
        try:
            deleted_comment = delete_comment(comment_id)
        except ForumV2RequestError:
            return Response(
                {"error": f"Comment does not exist with Id: {comment_id}"},
                status=status.HTTP_400_BAD_REQUEST,
            )
        return Response(deleted_comment, status=status.HTTP_200_OK)


class CreateThreadCommentAPIView(APIView):
    """
    API View to handle POST request for parent comments.
    """

    permission_classes = (AllowAny,)

    def post(self, request: Request, thread_id: str) -> Response:
        """
        Creates a new parent comment.

        Parameters:
            request (Request): The incoming request.
            thread_id: The ID of the thread for creating a comment on it.
        Body:
            body: The content of the comment.
            course_id: The Id of the respective course.
            user_id: The requesting user id.
            anonymous: anonymous flag(True or False).
            anonymous_to_peers: anonymous to peers flag(True or False).
        Response:
            The details of the comment that is created.
        """
        try:
            request_data = request.data
            comment = create_parent_comment(
                thread_id,
                request_data["body"],
                request_data["user_id"],
                request_data["course_id"],
                str_to_bool(request_data.get("anonymous", False)),
                str_to_bool(request_data.get("anonymous_to_peers", False)),
            )
        except ForumV2RequestError:
            return Response(
                {"error": f"Thread does not exist with Id: {thread_id}"},
                status=status.HTTP_400_BAD_REQUEST,
            )
        except ValueError as e:
            return Response(
                {"error": e},
                status=status.HTTP_400_BAD_REQUEST,
            )
        except ValidationError as e:
            return Response(
                {"error": e.detail},
                status=status.HTTP_400_BAD_REQUEST,
            )
        return Response(comment, status=status.HTTP_200_OK)



"""
Unit tests for the meilisearch search backend.
"""

from unittest.mock import patch, Mock

import search.meilisearch as m
from forum.search import meilisearch

TEST_ID = "abcd"
TEST_PK = m.id2pk(TEST_ID)


def test_create_document() -> None:
    assert {
        "id": TEST_ID,
        m.PRIMARY_KEY_FIELD_NAME: TEST_PK,
    } == meilisearch.create_document({}, TEST_ID)

    assert {
        "id": TEST_ID,
        m.PRIMARY_KEY_FIELD_NAME: TEST_PK,
        "body": "Somebody",
    } == meilisearch.create_document({"body": "Somebody"}, TEST_ID)

    assert {
        "id": TEST_ID,
        m.PRIMARY_KEY_FIELD_NAME: TEST_PK,
        "course_id": "some/course/id",
    } == meilisearch.create_document({"course_id": "some/course/id"}, TEST_ID)

    assert {
        "id": TEST_ID,
        m.PRIMARY_KEY_FIELD_NAME: TEST_PK,
    } == meilisearch.create_document(
        {"field_should_not_be_here": "some_value"}, TEST_ID
    )

    assert {
        "id": TEST_ID,
        m.PRIMARY_KEY_FIELD_NAME: TEST_PK,
        "body": "Somebody",
    } == meilisearch.create_document({"body": "<p>Somebody</p>"}, TEST_ID)


def test_index_document() -> None:
    backend = meilisearch.MeilisearchDocumentBackend()
    with patch.object(
        backend, "get_index", return_value=Mock(add_documents=Mock())
    ) as mock_get_index:
        backend.index_document(
            "my_index",
            TEST_ID,
            {
                "body": "<p>Some body</p>",
                "some other field": "some value",
            },
        )
        mock_get_index.assert_called_once_with("my_index")
        mock_get_index().add_documents.assert_called_once_with(
            [
                {
                    "id": TEST_ID,
                    "_pk": TEST_PK,
                    "body": "Some body",
                }
            ]
        )


def test_delete_document() -> None:
    backend = meilisearch.MeilisearchDocumentBackend()
    with patch.object(
        backend, "get_index", return_value=Mock(add_documents=Mock())
    ) as mock_get_index:
        backend.delete_document("my_index", TEST_ID)
        mock_get_index.assert_called_once_with("my_index")
        mock_get_index().delete_document.assert_called_once_with(TEST_PK)



"""Backend module for forum."""

from typing import Callable, Optional

from forum.backends.mongodb.api import MongoBackend
from forum.backends.mysql.api import MySQLBackend


def is_mysql_backend_enabled(course_id: str | None) -> bool:
    """
    Return True if mysql backend is enabled for the course.
    """
    try:
        # pylint: disable=import-outside-toplevel
        from forum.toggles import ENABLE_MYSQL_BACKEND
        from opaque_keys.edx.keys import CourseKey
    except ImportError:
        return True

    course_key: "CourseKey" | None = None
    if isinstance(course_id, CourseKey):
        course_key = course_id  # type: ignore[unreachable]
    elif isinstance(course_id, str):
        course_key = CourseKey.from_string(course_id)

    return ENABLE_MYSQL_BACKEND.is_enabled(course_key)


def get_backend(
    course_id: Optional[str] = None,
) -> Callable[[], MongoBackend | MySQLBackend]:
    """Return a factory function that lazily loads the backend API based on course_id."""

    def _get_backend() -> MongoBackend | MySQLBackend:
        if is_mysql_backend_enabled(course_id):
            return MySQLBackend()
        return MongoBackend()

    return _get_backend



"""Forum Flag API Views."""

from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from forum.api.flags import update_comment_flag, update_thread_flag
from forum.utils import ForumV2RequestError, str_to_bool


class CommentFlagAPIView(APIView):
    """
    API View for flagging/unflagging comments.

    Handles PUT requests to flag or unflag a comment.
    """

    permission_classes = (AllowAny,)

    def put(self, request: Request, comment_id: str, action: str) -> Response:
        """
        Flag or unflag a comment.

        Parameters:
        request (Request): The incoming request.
        comment_id (str): The ID of the comment to flag/unflag.
        action (str): The action to take (either "flag" or "unflag").

        Returns:
        Response: A response with the updated comment data.
        """
        request_data = request.data
        update_all = str_to_bool(request_data.get("all", False))
        user_id = request_data.get("user_id")
        try:
            serializer_data = update_comment_flag(
                comment_id, action, user_id, update_all
            )
            return Response(serializer_data, status=status.HTTP_200_OK)
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)


class ThreadFlagAPIView(APIView):
    """
    API View for flagging/unflagging threads.

    Handles PUT requests to flag or unflag a thread.
    """

    permission_classes = (AllowAny,)

    def put(self, request: Request, thread_id: str, action: str) -> Response:
        """
        Flag or unflag a thread.

        Parameters:
        request (Request): The incoming request.
        thread_id (str): The ID of the thread to flag/unflag.
        action (str): The action to take (either "flag" or "unflag").

        Returns:
        Response: A response with the updated thread data.
        """
        request_data = request.data
        update_all = str_to_bool(request_data.get("all", False))
        user_id = request_data.get("user_id")
        try:
            serializer_data = update_thread_flag(thread_id, action, user_id, update_all)
            return Response(serializer_data, status=status.HTTP_200_OK)
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)



"""Forum Views."""

from django.http import HttpRequest
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.views import APIView

from forum.utils import handle_proxy_requests


class ForumProxyAPIView(APIView):
    """
    A Proxy API View to Redirect All API requests to forum/cs_comments_service urls.
    """

    permission_classes = (AllowAny,)

    def post(self, request: HttpRequest, suffix: str) -> Response:
        """
        Catches post requests and sends them to forum/cs_comments_service post URLs.
        """
        response = handle_proxy_requests(request, suffix, "post")
        return Response(data=response.json(), status=response.status_code)

    def get(self, request: HttpRequest, suffix: str) -> Response:
        """
        Catches get requests and sends them to forum/cs_comments_service get URLs.
        """
        response = handle_proxy_requests(request, suffix, "get")
        if response.content:
            return Response(data=response.json(), status=response.status_code)
        else:
            return Response(data={}, status=response.status_code)

    def delete(self, request: HttpRequest, suffix: str) -> Response:
        """
        Catches delete requests and sends them to forum/cs_comments_service delete URLs.
        """
        response = handle_proxy_requests(request, suffix, "delete")
        return Response(data=response.json(), status=response.status_code)

    def put(self, request: HttpRequest, suffix: str) -> Response:
        """
        Catches put requests and sends them to forum/cs_comments_service put URLs.
        """
        response = handle_proxy_requests(request, suffix, "put")
        return Response(data=response.json(), status=response.status_code)



"""Users Serializers class."""

from typing import Any

from rest_framework import serializers


class UserSerializer(serializers.Serializer[Any]):
    """Serializer for users."""

    id = serializers.CharField(allow_null=True)
    username = serializers.CharField()
    email = serializers.CharField(allow_null=True)
    external_id = serializers.CharField()
    subscribed_thread_ids = serializers.ListField(
        child=serializers.CharField(), default=[]
    )
    subscribed_commentable_ids = serializers.ListField(
        child=serializers.CharField(), default=[]
    )
    subscribed_user_ids = serializers.ListField(
        child=serializers.CharField(), default=[]
    )
    follower_ids = serializers.ListField(child=serializers.CharField(), default=[])
    upvoted_ids = serializers.ListField(child=serializers.CharField(), default=[])
    downvoted_ids = serializers.ListField(child=serializers.CharField(), default=[])
    default_sort_key = serializers.CharField(allow_null=True)

    def create(self, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError

    def update(self, instance: Any, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError



# Generated by Django 4.2.16 on 2024-10-17 07:15

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("forum", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="readstate",
            name="course_id",
            field=models.CharField(max_length=255),
        ),
        migrations.AlterUniqueTogether(
            name="lastreadtime",
            unique_together={("read_state", "comment_thread")},
        ),
        migrations.CreateModel(
            name="MongoContent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("content_object_id", models.PositiveIntegerField(null=True)),
                ("mongo_id", models.CharField(max_length=50, unique=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
    ]



"""
Custom Serializer for the DataTime
"""

from datetime import datetime
from typing import Any

from rest_framework import serializers


class CustomDateTimeField(serializers.DateTimeField):
    """
    A custom DateTimeField serializer for formatting datetime objects in ISO 8601 format.
    """

    def to_representation(self, value: Any) -> str:
        """
        Convert the datetime object to an ISO 8601 formatted string with microseconds and UTC suffix.
        """
        # Check if the value is a datetime object
        if isinstance(value, datetime):
            # Format the datetime to ISO 8601 with microseconds and append 'Z'
            return f"{value:%Y-%m-%dT%H:%M}:{value.second:02}Z"

        # If the value is not a datetime object, fallback to the default representation
        return super().to_representation(value)



"""
This module contains the functions to update the flag status of a comment.
"""

from typing import Any, Optional

from forum.backend import get_backend
from forum.serializers.comment import CommentSerializer
from forum.serializers.thread import ThreadSerializer
from forum.utils import ForumV2RequestError


def update_comment_flag(
    comment_id: str,
    action: str,
    user_id: Optional[str] = None,
    update_all: Optional[bool] = False,
    course_id: Optional[str] = None,
) -> dict[str, Any]:
    """
    Update the flag status of a comment.

    Args:
        user_id (str): The ID of the user.
        comment_id (str): The ID of the comment.
        action (str): The action to perform ("flag" or "unflag").
        update_all (bool, optional): Whether to update all flags. Defaults to False.
    """
    backend = get_backend(course_id)()
    if not user_id:
        raise ForumV2RequestError("user_id not provided in params")
    user = backend.get_user(user_id)
    try:
        comment = backend.get_comment(comment_id)
    except ValueError as exc:
        raise ForumV2RequestError("User / Comment doesn't exist") from exc
    if not user or not comment:
        raise ForumV2RequestError("User / Comment doesn't exist")

    if action == "flag":
        updated_comment = backend.flag_as_abuse(
            user_id, comment_id, entity_type="Comment"
        )
    elif action == "unflag":
        if update_all:
            updated_comment = backend.un_flag_all_as_abuse(
                comment_id, entity_type="Comment"
            )
        else:
            updated_comment = backend.un_flag_as_abuse(
                user_id, comment_id, entity_type="Comment"
            )
    else:
        raise ForumV2RequestError("Invalid action")

    if updated_comment is None:
        raise ForumV2RequestError("Failed to update comment")

    context = {
        "id": updated_comment["_id"],
        **updated_comment,
        "user_id": user["_id"],
        "username": user["username"],
        "type": "comment",
        "comment_thread_id": str(updated_comment.get("comment_thread_id", None)),
    }
    return CommentSerializer(context, backend=backend).data


def update_thread_flag(
    thread_id: str,
    action: str,
    user_id: Optional[str] = None,
    update_all: Optional[bool] = False,
    course_id: Optional[str] = None,
) -> dict[str, Any]:
    """
    Update the flag status of a thread.

    Args:
        user_id (str): The ID of the user.
        thread_id (str): The ID of the thread.
        action (str): The action to perform ("flag" or "unflag").
        update_all (bool, optional): Whether to update all flags. Defaults to False.
    """
    backend = get_backend(course_id)()
    if not user_id:
        raise ForumV2RequestError("user_id not provided in params")
    user = backend.get_user(user_id)
    thread = backend.get_thread(thread_id)
    if not user or not thread:
        raise ForumV2RequestError("User / Thread doesn't exist")

    if action == "flag":
        updated_thread = backend.flag_as_abuse(
            user_id, thread_id, entity_type="CommentThread"
        )
    elif action == "unflag":
        if update_all:
            updated_thread = backend.un_flag_all_as_abuse(
                thread_id, entity_type="CommentThread"
            )
        else:
            updated_thread = backend.un_flag_as_abuse(
                user_id, thread_id, entity_type="CommentThread"
            )
    else:
        raise ForumV2RequestError("Invalid action")

    if updated_thread is None:
        raise ForumV2RequestError("Failed to update thread")

    context = {
        "id": str(updated_thread["_id"]),
        **updated_thread,
        "user_id": user["_id"],
        "username": user["username"],
        "type": "thread",
        "thread_id": str(updated_thread.get("comment_thread_id", None)),
    }
    return ThreadSerializer(context, backend=backend).data



"""
Vote Views
"""

from rest_framework import status
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from forum.api.votes import (
    delete_comment_vote,
    delete_thread_vote,
    update_comment_votes,
    update_thread_votes,
)
from forum.utils import ForumV2RequestError


class ThreadVoteView(APIView):
    """
    API view to handle voting on threads.

    Endpoint:
    PUT /forum/api/v2/threads/{thread_id}/votes/
    DELETE /forum/api/v2/threads/{thread_id}/votes/

    Example:
    PUT : /forum/api/v2/threads/66af33634a1e1f001b7ed57f/votes/
    Body::

        {
            "user_id": "4",
            "value": "up"
        }

    DELETE : /forum/api/v2/threads/66af33634a1e1f001b7ed57f/votes/
    Params::

        {
            "user_id": "4"
        }
    """

    def put(self, request: Request, thread_id: str) -> Response:
        """
        Handles the upvote or downvote on a thread.

        Args:
            request (Request): The incoming request object.
            thread_id (str): The ID of the thread to vote on.

        Returns:
            Response: The HTTP response with the result of the vote operation.
        """
        try:
            thread_response = update_thread_votes(
                thread_id, request.data["user_id"], request.data["value"]
            )
        except (ForumV2RequestError, KeyError) as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)

        return Response(thread_response, status=status.HTTP_200_OK)

    def delete(self, request: Request, thread_id: str) -> Response:
        """
        Handles removing a vote from a thread.

        Args:
            request (Request): The incoming request object.
            thread_id (str): The ID of the thread to remove the vote from.

        Returns:
            Response: The HTTP response with the result of the remove vote operation.
        """
        try:
            user_id = request.query_params.get("user_id", "")
            thread_response = delete_thread_vote(thread_id, user_id)
        except (ForumV2RequestError, KeyError) as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)

        return Response(thread_response, status=status.HTTP_200_OK)


class CommentVoteView(APIView):
    """
    API view to handle voting on comments.

    Endpoint:
    PUT /forum/api/v2/comments/{comment_id}/votes/
    DELETE /forum/api/v2/comments/{comment_id}/votes/

    Example:
    PUT : /forum/api/v2/comments/66af33634a1e1f001b7ed57f/votes/
    Body::

        {
            "user_id": "4",
            "value": "up"
        }

    DELETE : /forum/api/v2/comments/66af33634a1e1f001b7ed57f/votes/
    Params::

        {
            "user_id": "4"
        }
    """

    def put(self, request: Request, comment_id: str) -> Response:
        """
        Handles the upvote or downvote on a comment.

        Args:
            request (Request): The incoming request object.
            comment_id (str): The ID of the comment to vote on.

        Returns:
            Response: The HTTP response with the result of the vote operation.
        """
        try:
            comment_response = update_comment_votes(
                comment_id, request.data["user_id"], request.data["value"]
            )
        except (ForumV2RequestError, KeyError) as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)

        return Response(comment_response, status=status.HTTP_200_OK)

    def delete(self, request: Request, comment_id: str) -> Response:
        """
        Handles removing a vote from a comment.

        Args:
            request (Request): The incoming request object.
            comment_id (str): The ID of the comment to remove the vote from.

        Returns:
            Response: The HTTP response with the result of the remove vote operation.
        """
        try:
            user_id = request.query_params.get("user_id", "")
            comment_response = delete_comment_vote(comment_id, user_id)
        except (ForumV2RequestError, KeyError) as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)

        return Response(comment_response, status=status.HTTP_200_OK)



"""Forum Pin/Unpin thread API Views."""

import logging
from typing import Any

from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from forum.api import pin_thread, unpin_thread
from forum.utils import ForumV2RequestError

log = logging.getLogger(__name__)


class PinThreadAPIView(APIView):
    """
    API View to Pin thread.
    Handles PUT requests to pin a thread.
    """

    permission_classes = (AllowAny,)

    def put(self, request: Request, thread_id: str) -> Response:
        """
        Pins a thread.
        Parameters:
            request (Request): The incoming request.
            thread_id (str): The ID of the thread to pin.
        Body:
            user_id: requesting user's id
        Response:
            A response with the updated thread data.
        """
        try:
            thread_data: dict[str, Any] = pin_thread(
                request.data.get("user_id", ""), thread_id
            )
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)

        return Response(thread_data, status=status.HTTP_200_OK)


class UnpinThreadAPIView(APIView):
    """
    API View to Unpin a thread.
    Handles PUT requests to unpin a thread.
    """

    permission_classes = (AllowAny,)

    def put(self, request: Request, thread_id: str) -> Response:
        """
        Unpins a thread.
        Parameters:
            request (Request): The incoming request.
            thread_id (str): The ID of the thread to unpin.
        Body:
            user_id: requesting user's id
        Response:
            A response with the updated thread data.
        """
        try:
            thread_data: dict[str, Any] = unpin_thread(
                request.data.get("user_id", ""), thread_id
            )
        except ForumV2RequestError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)

        return Response(thread_data, status=status.HTTP_200_OK)



"""MySQL models for forum v2."""

from __future__ import annotations

from datetime import datetime
from typing import Any, Optional

from django.contrib.auth.models import User  # pylint: disable=E5142
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import QuerySet
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from forum.utils import validate_upvote_or_downvote


class ForumUser(models.Model):
    """Forum user model."""

    class Meta:
        app_label = "forum"

    user: models.OneToOneField[User, User] = models.OneToOneField(
        User, related_name="forum", on_delete=models.CASCADE
    )
    default_sort_key: models.CharField[str, str] = models.CharField(
        max_length=25, default="date"
    )

    def to_dict(self, course_id: Optional[str] = None) -> dict[str, Any]:
        """Return a dictionary representation of the model."""
        course_stats = CourseStat.objects.filter(user=self.user)
        read_states = ReadState.objects.filter(user=self.user)

        if course_id:
            course_stat = course_stats.filter(course_id=course_id).first()
        else:
            course_stat = None

        return {
            "_id": self.user.pk,
            "default_sort_key": self.default_sort_key,
            "external_id": self.user.pk,
            "username": self.user.username,
            "email": self.user.email,
            "course_stats": (
                course_stat.to_dict()
                if course_stat
                else [stat.to_dict() for stat in course_stats]
            ),
            "read_states": [state.to_dict() for state in read_states],
        }


class CourseStat(models.Model):
    """Course stats model."""

    course_id: models.CharField[str, str] = models.CharField(max_length=255)
    active_flags: models.IntegerField[int, int] = models.IntegerField(default=0)
    inactive_flags: models.IntegerField[int, int] = models.IntegerField(default=0)
    threads: models.IntegerField[int, int] = models.IntegerField(default=0)
    responses: models.IntegerField[int, int] = models.IntegerField(default=0)
    replies: models.IntegerField[int, int] = models.IntegerField(default=0)
    last_activity_at: models.DateTimeField[Optional[datetime], datetime] = (
        models.DateTimeField(default=None, null=True, blank=True)
    )
    user: models.ForeignKey[User, User] = models.ForeignKey(
        User, related_name="course_stats", on_delete=models.CASCADE
    )

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the model."""
        return {
            "_id": str(self.pk),
            "active_flags": self.active_flags,
            "inactive_flags": self.inactive_flags,
            "threads": self.threads,
            "responses": self.responses,
            "replies": self.replies,
            "course_id": self.course_id,
            "last_activity_at": self.last_activity_at,
        }

    class Meta:
        app_label = "forum"
        unique_together = ("user", "course_id")


class Content(models.Model):
    """Content model."""

    index_name = ""

    author: models.ForeignKey[User, User] = models.ForeignKey(
        User, on_delete=models.CASCADE
    )
    course_id: models.CharField[str, str] = models.CharField(max_length=255)
    body: models.TextField[str, str] = models.TextField()
    visible: models.BooleanField[bool, bool] = models.BooleanField(default=True)
    endorsed: models.BooleanField[bool, bool] = models.BooleanField(default=False)
    anonymous: models.BooleanField[bool, bool] = models.BooleanField(default=False)
    anonymous_to_peers: models.BooleanField[bool, bool] = models.BooleanField(
        default=False
    )
    group_id: models.PositiveIntegerField[int, int] = models.PositiveIntegerField(
        null=True
    )
    created_at: models.DateTimeField[datetime, datetime] = models.DateTimeField(
        auto_now_add=True
    )
    updated_at: models.DateTimeField[datetime, datetime] = models.DateTimeField(
        auto_now=True
    )
    uservote = GenericRelation(
        "UserVote",
        object_id_field="content_object_id",
        content_type_field="content_type",
    )

    @property
    def type(self) -> str:
        """Return the type of content as str."""
        return self._meta.object_name or ""

    @property
    def content_type(self) -> ContentType:
        """Return the type of content."""
        return ContentType.objects.get_for_model(self)

    @property
    def abuse_flaggers(self) -> list[int]:
        """Return a list of users who have flagged the content for abuse."""
        return list(
            AbuseFlagger.objects.filter(
                content_object_id=self.pk, content_type=self.content_type
            ).values_list("user_id", flat=True)
        )

    @property
    def historical_abuse_flaggers(self) -> list[int]:
        """Return a list of users who have historically flagged the content for abuse."""
        return list(
            HistoricalAbuseFlagger.objects.filter(
                content_object_id=self.pk, content_type=self.content_type
            ).values_list("user_id", flat=True)
        )

    @property
    def edit_history(self) -> QuerySet[EditHistory]:
        """Return a list of edit history for the content."""
        return EditHistory.objects.filter(
            content_object_id=self.pk, content_type=self.content_type
        )

    @property
    def votes(self) -> models.QuerySet[UserVote]:
        """Get all user vote query for content."""
        return UserVote.objects.filter(
            content_object_id=self.pk,
            content_type=self.content_type,
        )

    @property
    def get_votes(self) -> dict[str, Any]:
        """Get all user votes for content."""
        votes: dict[str, Any] = {
            "up": [],
            "down": [],
            "up_count": 0,
            "down_count": 0,
            "count": 0,
            "point": 0,
        }
        for vote in self.votes:
            if vote.vote == 1:
                votes["up"].append(vote.user.pk)
                votes["up_count"] += 1
            elif vote.vote == -1:
                votes["down"].append(vote.user.pk)
                votes["down_count"] += 1
            votes["point"] = votes["up_count"] - votes["down_count"]
            votes["count"] = votes["count"]
        return votes

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the content."""
        raise NotImplementedError

    class Meta:
        app_label = "forum"
        abstract = True


class CommentThread(Content):
    """Comment thread model."""

    index_name = "comment_threads"

    THREAD_TYPE_CHOICES = [
        ("question", "Question"),
        ("discussion", "Discussion"),
    ]

    CONTEXT_CHOICES = [
        ("course", "Course"),
        ("standalone", "Standalone"),
    ]

    title: models.CharField[str, str] = models.CharField(max_length=1024)
    thread_type: models.CharField[str, str] = models.CharField(
        max_length=50, choices=THREAD_TYPE_CHOICES, default="discussion"
    )
    context: models.CharField[str, str] = models.CharField(
        max_length=50, choices=CONTEXT_CHOICES, default="course"
    )
    closed: models.BooleanField[bool, bool] = models.BooleanField(default=False)
    pinned: models.BooleanField[Optional[bool], bool] = models.BooleanField(
        null=True, blank=True
    )
    last_activity_at: models.DateTimeField[Optional[datetime], datetime] = (
        models.DateTimeField(null=True, blank=True)
    )
    close_reason_code: models.CharField[Optional[str], str] = models.CharField(
        max_length=255, null=True, blank=True
    )
    closed_by: models.ForeignKey[User, User] = models.ForeignKey(
        User,
        related_name="threads_closed",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    commentable_id: models.CharField[str, str] = models.CharField(
        max_length=255,
        default=None,
        blank=True,
        null=True,
    )

    @property
    def comment_count(self) -> int:
        """Return the number of comments in the thread."""
        return Comment.objects.filter(comment_thread=self).count()

    @classmethod
    def get(cls, thread_id: str) -> CommentThread:
        """Get a comment thread model instance."""
        return cls.objects.get(pk=int(thread_id))

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the model."""
        edit_history = []
        for edit in self.edit_history.all():
            edit_history.append(
                {
                    "_id": str(edit.pk),
                    "original_body": edit.original_body,
                    "reason_code": edit.reason_code,
                    "editor_username": edit.editor.username,
                    "author_id": edit.editor.pk,
                    "created_at": edit.created_at,
                }
            )

        return {
            "_id": str(self.pk),
            "votes": self.get_votes,
            "visible": self.visible,
            "abuse_flaggers": [str(flagger) for flagger in self.abuse_flaggers],
            "historical_abuse_flaggers": [
                str(flagger) for flagger in self.historical_abuse_flaggers
            ],
            "thread_type": self.thread_type,
            "_type": "CommentThread",
            "commentable_id": self.commentable_id,
            "context": self.context,
            "comment_count": self.comment_count,
            "at_position_list": [],
            "pinned": self.pinned if self.pinned else False,
            "title": self.title,
            "body": self.body,
            "course_id": self.course_id,
            "anonymous": self.anonymous,
            "anonymous_to_peers": self.anonymous_to_peers,
            "closed": self.closed,
            "closed_by_id": str(self.closed_by.pk) if self.closed_by else None,
            "close_reason_code": self.close_reason_code,
            "author_id": str(self.author.pk),
            "author_username": self.author.username,
            "updated_at": self.updated_at,
            "created_at": self.created_at,
            "last_activity_at": self.last_activity_at,
            "edit_history": edit_history,
            "group_id": self.group_id,
        }

    def doc_to_hash(self) -> dict[str, Any]:
        """
        Converts the CommentThread model instance to a dictionary representation for Elasticsearch.
        """
        return {
            "id": str(self.pk),
            "title": self.title,
            "body": self.body,
            "created_at": self.created_at.isoformat() if self.created_at else None,
            "updated_at": self.updated_at.isoformat() if self.updated_at else None,
            "last_activity_at": (
                self.last_activity_at.isoformat() if self.last_activity_at else None
            ),
            "comment_count": self.comment_count,
            "votes_point": self.get_votes.get("point"),
            "context": self.context,
            "course_id": self.course_id,
            "commentable_id": self.commentable_id,
            "author_id": str(self.author.pk),
            "group_id": self.group_id,
            "thread_id": str(self.pk),
        }

    class Meta:
        app_label = "forum"
        indexes = [
            models.Index(fields=["context"]),
            models.Index(fields=["author"]),
            models.Index(fields=["author", "course_id"]),
            models.Index(fields=["course_id", "anonymous", "anonymous_to_peers"]),
            models.Index(
                fields=["author", "course_id", "anonymous", "anonymous_to_peers"]
            ),
        ]


class Comment(Content):
    """Comment model class"""

    index_name = "comments"

    endorsement: models.JSONField[dict[str, Any], dict[str, Any]] = models.JSONField(
        default=dict
    )
    sort_key: models.CharField[Optional[str], str] = models.CharField(
        max_length=255, null=True, blank=True
    )
    child_count: models.PositiveIntegerField[int, int] = models.PositiveIntegerField(
        default=0
    )
    retired_username: models.CharField[Optional[str], str] = models.CharField(
        max_length=255, null=True, blank=True
    )
    comment_thread: models.ForeignKey[CommentThread, CommentThread] = models.ForeignKey(
        CommentThread, on_delete=models.CASCADE
    )
    parent: models.ForeignKey[Comment, Comment] = models.ForeignKey(
        "self", on_delete=models.CASCADE, null=True, blank=True
    )
    depth: models.PositiveIntegerField[int, int] = models.PositiveIntegerField(
        default=0
    )

    def get_sort_key(self) -> str:
        """Get the sort key for the comment"""
        if self.parent:
            return f"{self.parent.pk}-{self.pk}"
        return str(self.pk)

    @staticmethod
    def get_list(**kwargs: Any) -> list[dict[str, Any]]:
        """
        Retrieves a list of all comments in the database based on provided filters.

        Args:
            kwargs: The filter arguments.

        Returns:
            A list of comments.
        """
        sort = kwargs.pop("sort", None)
        resp_skip = kwargs.pop("resp_skip", 0)
        resp_limit = kwargs.pop("resp_limit", None)
        comments = Comment.objects.filter(**kwargs)
        result = []
        if sort:
            if sort == 1:
                result = sorted(
                    comments, key=lambda x: (x.sort_key is None, x.sort_key or "")
                )
            elif sort == -1:
                result = sorted(
                    comments,
                    key=lambda x: (x.sort_key is None, x.sort_key or ""),
                    reverse=True,
                )

        paginated_comments = result or list(comments)

        # Apply pagination if resp_limit is provided
        if resp_limit is not None:
            resp_end = resp_skip + resp_limit
            paginated_comments = result[resp_skip:resp_end]
        elif resp_skip:  # If resp_limit is None but resp_skip is provided
            paginated_comments = result[resp_skip:]

        return [content.to_dict() for content in paginated_comments]

    @staticmethod
    def get_list_total_count(**kwargs: Any) -> int:
        """
        Retrieves the total count of comments in the database based on provided filters.

        Args:
            kwargs: The filter arguments to apply when counting comments.

        Returns:
            The total number of comments matching the provided filters.
        """
        kwargs.pop("sort", None)
        kwargs.pop("resp_skip", 0)
        kwargs.pop("resp_limit", None)
        return Comment.objects.filter(**kwargs).count()

    def get_parent_ids(self) -> list[str]:
        """Return a list of all parent IDs of a comment."""
        parent_ids = []
        current_comment = self
        while current_comment.parent:
            parent_ids.append(str(current_comment.parent.pk))
            current_comment = current_comment.parent
        return parent_ids

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the model."""
        edit_history = []
        for edit in self.edit_history.all():
            edit_history.append(
                {
                    "_id": str(edit.pk),
                    "original_body": edit.original_body,
                    "reason_code": edit.reason_code,
                    "editor_username": edit.editor.username,
                    "author_id": edit.editor.pk,
                    "created_at": edit.created_at,
                }
            )

        endorsement = {
            "user_id": self.endorsement.get("user_id") if self.endorsement else None,
            "time": self.endorsement.get("time") if self.endorsement else None,
        }

        data = {
            "_id": str(self.pk),
            "votes": self.get_votes,
            "visible": self.visible,
            "abuse_flaggers": [str(flagger) for flagger in self.abuse_flaggers],
            "historical_abuse_flaggers": [
                str(flagger) for flagger in self.historical_abuse_flaggers
            ],
            "parent_ids": self.get_parent_ids(),
            "parent_id": str(self.parent.pk) if self.parent else "None",
            "at_position_list": [],
            "body": self.body,
            "course_id": self.course_id,
            "_type": "Comment",
            "endorsed": self.endorsed,
            "anonymous": self.anonymous,
            "anonymous_to_peers": self.anonymous_to_peers,
            "author_id": str(self.author.pk),
            "comment_thread_id": str(self.comment_thread.pk),
            "child_count": self.child_count,
            "author_username": self.author.username,
            "sk": str(self.pk),
            "updated_at": self.updated_at,
            "created_at": self.created_at,
            "endorsement": endorsement if self.endorsement else None,
        }
        if edit_history:
            data["edit_history"] = edit_history

        return data

    @classmethod
    def get(cls, comment_id: str) -> Comment:
        """Get a comment model instance."""
        return cls.objects.get(pk=int(comment_id))

    def doc_to_hash(self) -> dict[str, Any]:
        """
        Converts the Comment model instance to a dictionary representation for Elasticsearch.
        """
        return {
            "body": self.body,
            "course_id": self.course_id,
            "comment_thread_id": self.comment_thread.pk,
            "commentable_id": None,
            "group_id": self.group_id,
            "context": "course",
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
            "title": None,
        }

    class Meta:
        app_label = "forum"
        indexes = [
            models.Index(fields=["author", "course_id"]),
            models.Index(fields=["comment_thread", "author", "created_at"]),
            models.Index(fields=["comment_thread", "endorsed"]),
            models.Index(fields=["course_id", "parent", "endorsed"]),
            models.Index(fields=["course_id", "anonymous", "anonymous_to_peers"]),
            models.Index(
                fields=["author", "course_id", "anonymous", "anonymous_to_peers"]
            ),
        ]


class EditHistory(models.Model):
    """Edit history model class"""

    DISCUSSION_MODERATION_EDIT_REASON_CODES = [
        ("grammar-spelling", _("Has grammar / spelling issues")),
        ("needs-clarity", _("Content needs clarity")),
        ("academic-integrity", _("Has academic integrity concern")),
        ("inappropriate-language", _("Has inappropriate language")),
        ("format-change", _("Formatting changes needed")),
        ("post-type-change", _("Post type needs change")),
        ("contains-pii", _("Contains personally identifiable information")),
        ("violates-guidelines", _("Violates community guidelines")),
    ]

    reason_code: models.CharField[Optional[str], str] = models.CharField(
        max_length=100,
        blank=True,
        null=True,
        choices=DISCUSSION_MODERATION_EDIT_REASON_CODES,
    )
    original_body: models.TextField[str, str] = models.TextField()
    editor: models.ForeignKey[User, User] = models.ForeignKey(
        User, on_delete=models.CASCADE
    )
    created_at: models.DateTimeField[datetime, datetime] = models.DateTimeField(
        auto_now_add=True
    )
    content_type: models.ForeignKey[ContentType] = models.ForeignKey(
        ContentType, on_delete=models.CASCADE
    )
    content_object_id: models.PositiveIntegerField[int, int] = (
        models.PositiveIntegerField()
    )
    content: GenericForeignKey = GenericForeignKey("content_type", "content_object_id")

    class Meta:
        app_label = "forum"
        indexes = [
            models.Index(fields=["editor"]),
            models.Index(fields=["content_type", "content_object_id"]),
            models.Index(fields=["created_at"]),
        ]


class AbuseFlagger(models.Model):
    """Abuse flagger model class"""

    content_type: models.ForeignKey[ContentType] = models.ForeignKey(
        ContentType, on_delete=models.CASCADE
    )
    content_object_id: models.PositiveIntegerField[int, int] = (
        models.PositiveIntegerField()
    )
    content: GenericForeignKey = GenericForeignKey("content_type", "content_object_id")
    user: models.ForeignKey[User, User] = models.ForeignKey(
        User, on_delete=models.CASCADE
    )
    flagged_at: models.DateTimeField[datetime, datetime] = models.DateTimeField(
        default=timezone.now
    )

    class Meta:
        app_label = "forum"
        unique_together = ("user", "content_type", "content_object_id")
        indexes = [
            models.Index(fields=["content_type", "content_object_id"]),
            models.Index(fields=["user", "content_type", "content_object_id"]),
        ]


class HistoricalAbuseFlagger(models.Model):
    """Historical abuse flagger model class"""

    content_type: models.ForeignKey[ContentType] = models.ForeignKey(
        ContentType, on_delete=models.CASCADE
    )
    content_object_id: models.PositiveIntegerField[int, int] = (
        models.PositiveIntegerField()
    )
    content: GenericForeignKey = GenericForeignKey("content_type", "content_object_id")
    user: models.ForeignKey[User, User] = models.ForeignKey(
        User, on_delete=models.CASCADE
    )
    flagged_at: models.DateTimeField[datetime, datetime] = models.DateTimeField(
        default=timezone.now
    )

    class Meta:
        app_label = "forum"
        unique_together = ("user", "content_type", "content_object_id")
        indexes = [
            models.Index(fields=["content_type", "content_object_id"]),
            models.Index(fields=["user", "content_type", "content_object_id"]),
        ]


class ReadState(models.Model):
    """Read state model."""

    course_id: models.CharField[str, str] = models.CharField(max_length=255)
    user: models.ForeignKey[User, User] = models.ForeignKey(
        User, related_name="read_states", on_delete=models.CASCADE
    )
    last_read_times: models.QuerySet[LastReadTime]

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the model."""
        last_read_times = {}
        for last_read_time in self.last_read_times.all():
            last_read_times[str(last_read_time.comment_thread.pk)] = (
                last_read_time.timestamp
            )
        return {
            "_id": str(self.pk),
            "last_read_times": last_read_times,
            "course_id": self.course_id,
        }

    class Meta:
        app_label = "forum"
        unique_together = ("course_id", "user")
        indexes = [
            models.Index(fields=["user", "course_id"]),
        ]


class LastReadTime(models.Model):
    """Last read time model."""

    read_state: models.ForeignKey[ReadState] = models.ForeignKey(
        ReadState, related_name="last_read_times", on_delete=models.CASCADE
    )
    comment_thread: models.ForeignKey[CommentThread, CommentThread] = models.ForeignKey(
        CommentThread, on_delete=models.CASCADE
    )
    timestamp: models.DateTimeField[datetime, datetime] = models.DateTimeField()

    class Meta:
        app_label = "forum"
        unique_together = ("read_state", "comment_thread")
        indexes = [
            models.Index(fields=["read_state", "timestamp"]),
            models.Index(fields=["comment_thread"]),
        ]


class UserVote(models.Model):
    """User votes model class"""

    user: models.ForeignKey[User, User] = models.ForeignKey(
        User, on_delete=models.CASCADE
    )
    content_type: models.ForeignKey[ContentType] = models.ForeignKey(
        ContentType, on_delete=models.CASCADE
    )
    content_object_id: models.PositiveIntegerField[int, int] = (
        models.PositiveIntegerField()
    )
    content: GenericForeignKey = GenericForeignKey("content_type", "content_object_id")
    vote: models.IntegerField[int, int] = models.IntegerField(
        validators=[validate_upvote_or_downvote]
    )

    class Meta:
        app_label = "forum"
        unique_together = ("user", "content_type", "content_object_id")
        indexes = [
            models.Index(fields=["vote"]),
            models.Index(fields=["user", "vote"]),
            models.Index(fields=["content_type", "content_object_id"]),
        ]


class Subscription(models.Model):
    """Subscription model class"""

    subscriber: models.ForeignKey[User, User] = models.ForeignKey(
        User, on_delete=models.CASCADE
    )
    source_content_type: models.ForeignKey[ContentType, ContentType] = (
        models.ForeignKey(ContentType, on_delete=models.CASCADE)
    )
    source_object_id: models.PositiveIntegerField[int, int] = (
        models.PositiveIntegerField()
    )
    source: GenericForeignKey = GenericForeignKey(
        "source_content_type", "source_object_id"
    )
    created_at: models.DateTimeField[datetime, datetime] = models.DateTimeField(
        auto_now_add=True
    )
    updated_at: models.DateTimeField[datetime, datetime] = models.DateTimeField(
        auto_now=True
    )

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the model."""
        return {
            "_id": str(self.pk),
            "subscriber_id": str(self.subscriber.pk),
            "source_id": str(self.source_object_id),
            "source_type": self.source_content_type.model,
            "updated_at": self.updated_at,
            "created_at": self.created_at,
        }

    class Meta:
        app_label = "forum"
        unique_together = ("subscriber", "source_content_type", "source_object_id")
        indexes = [
            models.Index(fields=["subscriber"]),
            models.Index(
                fields=["subscriber", "source_object_id", "source_content_type"]
            ),
            models.Index(fields=["subscriber", "source_content_type"]),
            models.Index(fields=["source_object_id", "source_content_type"]),
        ]


class MongoContent(models.Model):
    """MongoContent model class."""

    content_type: models.ForeignKey[ContentType] = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True
    )
    content_object_id: models.PositiveIntegerField[int, int] = (
        models.PositiveIntegerField(null=True)
    )
    content: GenericForeignKey = GenericForeignKey("content_type", "content_object_id")
    mongo_id: models.CharField[str, str] = models.CharField(max_length=50, unique=True)

    class Meta:
        app_label = "forum"



"""Serializer class for content collection."""

from typing import Any

from rest_framework import serializers

from forum.serializers.custom_datetime import CustomDateTimeField
from forum.serializers.votes import VoteSummarySerializer


class EditHistorySerializer(serializers.Serializer[dict[str, Any]]):
    """
    Serializer for handling edit history of a post or comment

    Attributes:
        original_body (str): The original content of the post or comment before the edit.
        reason_code (str): The code representing the reason for editing the post or comment.
        editor_username (str): The username of the person who made the edit.
        created_at (datetime): The timestamp of when the edit was made.
    """

    original_body = serializers.CharField()
    reason_code = serializers.CharField(allow_null=True, default=None)
    editor_username = serializers.CharField()
    created_at = CustomDateTimeField()

    def create(self, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError

    def update(self, instance: Any, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError


class ContentSerializer(serializers.Serializer[dict[str, Any]]):
    """
    Serializer for handling the content of a post or comment.

    Attributes:
        id (str): The ID of the Content.
        body (str): The main content text.
        course_id (str): The ID of the related course.
        anonymous (bool): Whether the content is posted anonymously.
        anonymous_to_peers (bool): Whether the content is anonymous to peers.
        created_at (datetime): The timestamp when the content was created.
        updated_at (datetime): The timestamp when the content was last updated.
        at_position_list (list): A list of positions where @mentions occur.
        user_id (str): The ID of the user who created the content.
        username (str): The username of the content creator.
        commentable_id (str): The ID of the entity the content is related to (e.g., course).
        votes (VoteSummarySerializer): A summary of votes on the content.
        abuse_flaggers (list): A list of user IDs who flagged the content as abusive.
        historical_abuse_flaggers (list): A list of user IDs who historically flagged the content as abusive.
        edit_history (list): A list of previous versions of the content.
        closed (bool): Whether the content is closed for further interactions.
        type (str): The type of content (e.g., "post", "comment").
    """

    id = serializers.CharField(source="_id")
    body = serializers.CharField()
    course_id = serializers.CharField()
    anonymous = serializers.BooleanField(default=False)
    anonymous_to_peers = serializers.BooleanField(default=False)
    created_at = CustomDateTimeField(allow_null=True)
    updated_at = CustomDateTimeField(allow_null=True)
    at_position_list = serializers.ListField(default=[])
    user_id = serializers.CharField(source="author_id")
    username = serializers.CharField(source="author_username", allow_null=True)
    commentable_id = serializers.CharField(default="course")
    votes = VoteSummarySerializer()
    abuse_flaggers = serializers.ListField(child=serializers.CharField(), default=[])
    historical_abuse_flaggers = serializers.ListField(
        child=serializers.CharField(), default=[]
    )
    edit_history = EditHistorySerializer(default=[], many=True)
    closed = serializers.BooleanField(default=False)
    type = serializers.CharField()

    def create(self, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError

    def update(self, instance: Any, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError



# Generated by Django 4.2.14 on 2024-10-10 15:57

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import forum.utils


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="CommentThread",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("course_id", models.CharField(max_length=255)),
                ("body", models.TextField()),
                ("visible", models.BooleanField(default=True)),
                ("endorsed", models.BooleanField(default=False)),
                ("anonymous", models.BooleanField(default=False)),
                ("anonymous_to_peers", models.BooleanField(default=False)),
                ("group_id", models.PositiveIntegerField(null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("title", models.CharField(max_length=255)),
                (
                    "thread_type",
                    models.CharField(
                        choices=[
                            ("question", "Question"),
                            ("discussion", "Discussion"),
                        ],
                        default="discussion",
                        max_length=50,
                    ),
                ),
                (
                    "context",
                    models.CharField(
                        choices=[("course", "Course"), ("standalone", "Standalone")],
                        default="course",
                        max_length=50,
                    ),
                ),
                ("closed", models.BooleanField(default=False)),
                ("pinned", models.BooleanField(blank=True, null=True)),
                ("last_activity_at", models.DateTimeField(blank=True, null=True)),
                (
                    "close_reason_code",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                (
                    "commentable_id",
                    models.CharField(
                        blank=True, default=None, max_length=255, null=True
                    ),
                ),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "closed_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="threads_closed",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ReadState",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("course_id", models.CharField(max_length=255, unique=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="read_states",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="LastReadTime",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField()),
                (
                    "comment_thread",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="forum.commentthread",
                    ),
                ),
                (
                    "read_state",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="last_read_times",
                        to="forum.readstate",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="HistoricalAbuseFlagger",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("content_object_id", models.PositiveIntegerField()),
                ("flagged_at", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ForumUser",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("default_sort_key", models.CharField(default="date", max_length=25)),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="forum",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EditHistory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "reason_code",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("grammar-spelling", "Has grammar / spelling issues"),
                            ("needs-clarity", "Content needs clarity"),
                            ("academic-integrity", "Has academic integrity concern"),
                            ("inappropriate-language", "Has inappropriate language"),
                            ("format-change", "Formatting changes needed"),
                            ("post-type-change", "Post type needs change"),
                            (
                                "contains-pii",
                                "Contains personally identifiable information",
                            ),
                            ("violates-guidelines", "Violates community guidelines"),
                        ],
                        max_length=100,
                        null=True,
                    ),
                ),
                ("original_body", models.TextField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("content_object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "editor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="CourseStat",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("course_id", models.CharField(max_length=255)),
                ("active_flags", models.IntegerField(default=0)),
                ("inactive_flags", models.IntegerField(default=0)),
                ("threads", models.IntegerField(default=0)),
                ("responses", models.IntegerField(default=0)),
                ("replies", models.IntegerField(default=0)),
                (
                    "last_activity_at",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="course_stats",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Comment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("course_id", models.CharField(max_length=255)),
                ("body", models.TextField()),
                ("visible", models.BooleanField(default=True)),
                ("endorsed", models.BooleanField(default=False)),
                ("anonymous", models.BooleanField(default=False)),
                ("anonymous_to_peers", models.BooleanField(default=False)),
                ("group_id", models.PositiveIntegerField(null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("endorsement", models.JSONField(default=dict)),
                ("sort_key", models.CharField(blank=True, max_length=255, null=True)),
                ("child_count", models.PositiveIntegerField(default=0)),
                (
                    "retired_username",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("depth", models.PositiveIntegerField(default=0)),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "comment_thread",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="forum.commentthread",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="forum.comment",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AbuseFlagger",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("content_object_id", models.PositiveIntegerField()),
                ("flagged_at", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UserVote",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("content_object_id", models.PositiveIntegerField()),
                (
                    "vote",
                    models.IntegerField(
                        validators=[forum.utils.validate_upvote_or_downvote]
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "indexes": [
                    models.Index(fields=["vote"], name="forum_userv_vote_2ccbcf_idx"),
                    models.Index(
                        fields=["user", "vote"], name="forum_userv_user_id_7ab0c3_idx"
                    ),
                    models.Index(
                        fields=["content_type", "content_object_id"],
                        name="forum_userv_content_55d9ad_idx",
                    ),
                ],
                "unique_together": {("user", "content_type", "content_object_id")},
            },
        ),
        migrations.CreateModel(
            name="Subscription",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("source_object_id", models.PositiveIntegerField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "source_content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "subscriber",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "indexes": [
                    models.Index(
                        fields=["subscriber"], name="forum_subsc_subscri_34fd6c_idx"
                    ),
                    models.Index(
                        fields=[
                            "subscriber",
                            "source_object_id",
                            "source_content_type",
                        ],
                        name="forum_subsc_subscri_77168d_idx",
                    ),
                    models.Index(
                        fields=["subscriber", "source_content_type"],
                        name="forum_subsc_subscri_a64d92_idx",
                    ),
                    models.Index(
                        fields=["source_object_id", "source_content_type"],
                        name="forum_subsc_source__ef74be_idx",
                    ),
                ],
                "unique_together": {
                    ("subscriber", "source_content_type", "source_object_id")
                },
            },
        ),
        migrations.AddIndex(
            model_name="readstate",
            index=models.Index(
                fields=["user", "course_id"], name="forum_reads_user_id_30cd0a_idx"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="readstate",
            unique_together={("course_id", "user")},
        ),
        migrations.AddIndex(
            model_name="lastreadtime",
            index=models.Index(
                fields=["read_state", "timestamp"],
                name="forum_lastr_read_st_67a4f9_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="lastreadtime",
            index=models.Index(
                fields=["comment_thread"], name="forum_lastr_comment_652102_idx"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="lastreadtime",
            unique_together={("read_state", "timestamp")},
        ),
        migrations.AddIndex(
            model_name="historicalabuseflagger",
            index=models.Index(
                fields=["content_type", "content_object_id"],
                name="forum_histo_content_e4456d_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="historicalabuseflagger",
            index=models.Index(
                fields=["user", "content_type", "content_object_id"],
                name="forum_histo_user_id_761c88_idx",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="historicalabuseflagger",
            unique_together={("user", "content_type", "content_object_id")},
        ),
        migrations.AddIndex(
            model_name="edithistory",
            index=models.Index(
                fields=["editor"], name="forum_edith_editor__159ea4_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="edithistory",
            index=models.Index(
                fields=["content_type", "content_object_id"],
                name="forum_edith_content_3e9471_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="edithistory",
            index=models.Index(
                fields=["created_at"], name="forum_edith_created_67bba8_idx"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="coursestat",
            unique_together={("user", "course_id")},
        ),
        migrations.AddIndex(
            model_name="commentthread",
            index=models.Index(
                fields=["context"], name="forum_comme_context_cfe3d0_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="commentthread",
            index=models.Index(
                fields=["author"], name="forum_comme_author__04774e_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="commentthread",
            index=models.Index(
                fields=["author", "course_id"], name="forum_comme_author__e6f558_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="commentthread",
            index=models.Index(
                fields=["course_id", "anonymous", "anonymous_to_peers"],
                name="forum_comme_course__c529cf_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="commentthread",
            index=models.Index(
                fields=["author", "course_id", "anonymous", "anonymous_to_peers"],
                name="forum_comme_author__8df595_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="comment",
            index=models.Index(
                fields=["author", "course_id"], name="forum_comme_author__257002_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="comment",
            index=models.Index(
                fields=["comment_thread", "author", "created_at"],
                name="forum_comme_comment_5e5849_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="comment",
            index=models.Index(
                fields=["comment_thread", "endorsed"],
                name="forum_comme_comment_4f1303_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="comment",
            index=models.Index(
                fields=["course_id", "parent", "endorsed"],
                name="forum_comme_course__333a3b_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="comment",
            index=models.Index(
                fields=["course_id", "anonymous", "anonymous_to_peers"],
                name="forum_comme_course__fc7371_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="comment",
            index=models.Index(
                fields=["author", "course_id", "anonymous", "anonymous_to_peers"],
                name="forum_comme_author__084b35_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="abuseflagger",
            index=models.Index(
                fields=["content_type", "content_object_id"],
                name="forum_abuse_content_b190f5_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="abuseflagger",
            index=models.Index(
                fields=["user", "content_type", "content_object_id"],
                name="forum_abuse_user_id_8f5d27_idx",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="abuseflagger",
            unique_together={("user", "content_type", "content_object_id")},
        ),
    ]



"""Management command for deleting unused indices"""

from django.core.management.base import BaseCommand

from forum.search import get_index_search_backend


class Command(BaseCommand):
    """
    Django management command for the deletion of unused search indices.
    """

    help = (
        "Delete all Elasticsearch indices that are not the latest for each model type."
    )

    def handle(self, *args: list[str], **kwargs: dict[str, str]) -> None:
        """
        Handles the execution of the delete_unused_forum_indices command.

        This command deletes all Elasticsearch indices that are not the latest for each model type.
        """
        search_backend = get_index_search_backend()
        indices_deleted_count = search_backend.delete_unused_indices()
        self.stdout.write(
            self.style.SUCCESS(
                f"{indices_deleted_count} unused indices deleted successfully."
            )
        )



"""Migration commands helper methods."""

from typing import Any

from django.contrib.auth.models import User  # pylint: disable=E5142
from django.core.management.base import OutputWrapper
from django.utils import timezone
from pymongo.collection import Collection
from pymongo.database import Database

from forum.models import (
    AbuseFlagger,
    Comment,
    CommentThread,
    CourseStat,
    EditHistory,
    ForumUser,
    HistoricalAbuseFlagger,
    LastReadTime,
    MongoContent,
    ReadState,
    Subscription,
    UserVote,
)
from forum.utils import make_aware, get_trunc_title


def get_all_course_ids(db: Database[dict[str, Any]]) -> list[str]:
    """Get all course IDs from MongoDB."""
    return db.contents.distinct("course_id")


def migrate_users(db: Database[dict[str, Any]], course_id: str) -> None:
    """Migrate users from MongoDB to MySQL."""
    users = db.users.find({"course_stats.course_id": course_id})
    for user_data in users:
        user, _ = User.objects.get_or_create(
            id=int(user_data["_id"]), defaults={"username": user_data["username"]}
        )

        _, _ = ForumUser.objects.get_or_create(
            user=user,
            defaults={"default_sort_key": user_data.get("default_sort_key", "date")},
        )

        for stat in user_data.get("course_stats", []):
            last_activity_at = stat.get("last_activity_at")
            if last_activity_at:
                last_activity_at = make_aware(last_activity_at)
            if stat["course_id"] == course_id:
                CourseStat.objects.update_or_create(
                    user=user,
                    course_id=course_id,
                    defaults={
                        "active_flags": stat.get("active_flags", 0),
                        "inactive_flags": stat.get("inactive_flags", 0),
                        "threads": stat.get("threads", 0),
                        "responses": stat.get("responses", 0),
                        "replies": stat.get("replies", 0),
                        "last_activity_at": last_activity_at,
                    },
                )


def migrate_content(db: Database[dict[str, Any]], course_id: str) -> None:
    """Migrate content from MongoDB to MySQL."""
    contents = db.contents.find({"course_id": course_id})
    for content in contents:
        if content["_type"] == "CommentThread":
            create_or_update_thread(content)
        elif content["_type"] == "Comment":
            create_or_update_comment(content)

        migrate_subscriptions(db, content["_id"])


def create_or_update_thread(thread_data: dict[str, Any]) -> None:
    """Create or update a thread."""
    author = User.objects.get(id=int(thread_data["author_id"]))
    mongo_thread_id = str(thread_data["_id"])
    mongo_content, _ = MongoContent.objects.get_or_create(
        mongo_id=mongo_thread_id,
    )
    if not mongo_content.content_object_id:
        thread = CommentThread.objects.create(
            author=author,
            course_id=thread_data["course_id"],
            title=get_trunc_title(thread_data.get("title", "")),
            body=thread_data["body"],
            thread_type=thread_data.get("thread_type", "discussion"),
            context=thread_data.get("context", "course"),
            anonymous=thread_data.get("anonymous", False),
            anonymous_to_peers=thread_data.get("anonymous_to_peers", False),
            closed=thread_data.get("closed", False),
            pinned=thread_data.get("pinned"),
            created_at=make_aware(thread_data["created_at"]),
            updated_at=make_aware(thread_data["updated_at"]),
            last_activity_at=make_aware(thread_data["last_activity_at"]),
            commentable_id=thread_data.get("commentable_id"),
        )
        mongo_content.content_object_id = thread.pk
        mongo_content.content_type = thread.content_type
        mongo_content.save()
    else:
        thread = CommentThread.objects.get(pk=mongo_content.content_object_id)

    create_or_update_edit_history(thread_data)
    create_or_update_abuse_flaggers(thread_data)
    create_votes(thread, thread_data.get("votes", {}))


def create_or_update_comment(comment_data: dict[str, Any]) -> None:
    """Create or update a comment."""
    author = User.objects.get(id=int(comment_data["author_id"]))
    mongo_thread_id = str(comment_data["comment_thread_id"])
    mongo_thread = MongoContent.objects.get(mongo_id=mongo_thread_id)
    thread = CommentThread.objects.get(pk=mongo_thread.content_object_id)
    parent = None
    if "parent_id" in comment_data and comment_data["parent_id"] != "None":
        parent_id = str(comment_data["parent_id"])
        mongo_parent_comment = MongoContent.objects.get(mongo_id=parent_id)
        parent = Comment.objects.filter(
            id=mongo_parent_comment.content_object_id
        ).first()

    mongo_comment, _ = MongoContent.objects.get_or_create(
        mongo_id=str(comment_data["_id"])
    )
    if not mongo_comment.content_object_id:
        comment = Comment.objects.create(
            author=author,
            comment_thread=thread,
            parent=parent,
            course_id=comment_data["course_id"],
            body=comment_data["body"],
            anonymous=comment_data.get("anonymous", False),
            anonymous_to_peers=comment_data.get("anonymous_to_peers", False),
            endorsed=comment_data.get("endorsed", False),
            child_count=comment_data.get("child_count", 0),
            created_at=make_aware(comment_data["created_at"]),
            updated_at=make_aware(comment_data["updated_at"]),
            depth=1 if parent else 0,
        )
        mongo_comment.content_object_id = comment.pk
        mongo_comment.content_type = comment.content_type
        mongo_comment.save()
        sort_key = f"{parent.pk}-{comment.pk}" if parent else f"{comment.pk}"
        comment.sort_key = sort_key
        comment.save()
    else:
        comment = Comment.objects.get(pk=mongo_comment.content_object_id)

    create_or_update_edit_history(comment_data)
    create_or_update_abuse_flaggers(comment_data)
    create_votes(comment, comment_data.get("votes", {}))


def create_votes(content: CommentThread | Comment, votes_data: dict[str, Any]) -> None:
    """Create or update votes for a content."""
    for vote_type in ["up", "down"]:
        for user_id in votes_data.get(vote_type, []):
            user = User.objects.get(pk=int(user_id))
            UserVote.objects.update_or_create(
                user=user,
                content_type=content.content_type,
                content_object_id=content.pk,
                defaults={"vote": 1 if vote_type == "up" else -1},
            )


def create_or_update_edit_history(content: dict[str, Any]) -> None:
    """Create or update edit history for a content."""
    edit_history = content.get("edit_history", [])
    content_type = CommentThread if content["_type"] == "CommentThread" else Comment
    mongo_content = MongoContent.objects.get(mongo_id=str(content["_id"]))
    content_object = content_type.objects.get(pk=mongo_content.content_object_id)
    for edit in edit_history:
        EditHistory.objects.get_or_create(
            content_object_id=content_object.pk,
            content_type=content_object.content_type,
            created_at=edit["created_at"],
            editor=User.objects.get(pk=int(edit["author_id"])),
            defaults={
                "original_body": edit["original_body"],
                "reason_code": edit["reason_code"],
            },
        )


def create_or_update_abuse_flaggers(content: dict[str, Any]) -> None:
    """Create or update abuse flaggers for content."""
    content_type = CommentThread if content["_type"] == "CommentThread" else Comment
    mongo_content = MongoContent.objects.get(mongo_id=str(content["_id"]))
    content_object = content_type.objects.get(pk=mongo_content.content_object_id)
    for user_id in content["abuse_flaggers"]:
        user = User.objects.get(pk=int(user_id))
        AbuseFlagger.objects.update_or_create(
            user=user,
            content_type=content_object.content_type,
            content_object_id=content_object.pk,
            defaults={
                "flagged_at": timezone.now(),
            },
        )
    for user_id in content["historical_abuse_flaggers"]:
        user = User.objects.get(pk=int(user_id))
        HistoricalAbuseFlagger.objects.update_or_create(
            user=user,
            content_type=content_object.content_type,
            content_object_id=content_object.pk,
            defaults={
                "flagged_at": timezone.now(),
            },
        )


def migrate_subscriptions(db: Database[dict[str, Any]], content_id: str) -> None:
    """Migrate subscriptions from mongo to mysql."""
    subscriptions = db.subscriptions.find({"source_id": str(content_id)})
    for sub in subscriptions:
        user = User.objects.get(id=int(sub["subscriber_id"]))
        content_type = (
            CommentThread if sub["source_type"] == "CommentThread" else Comment
        )
        mongo_content = MongoContent.objects.get(mongo_id=str(content_id))
        content = content_type.objects.get(pk=mongo_content.content_object_id)
        if content:
            Subscription.objects.update_or_create(
                subscriber=user,
                source_content_type=content.content_type,
                source_object_id=content.pk,
                defaults={
                    "created_at": sub.get("created_at", timezone.now()),
                    "updated_at": sub.get("updated_at", timezone.now()),
                },
            )


def migrate_read_states(db: Database[dict[str, Any]], course_id: str) -> None:
    """Migrate read states from mongo to mysql."""
    users = db.users.find({"course_stats.course_id": course_id})
    for user_data in users:
        user = User.objects.filter(id=int(user_data["_id"])).first()
        if not user:
            continue

        for read_state in user_data.get("read_states", []):
            if read_state["course_id"] == course_id:
                rs, _ = ReadState.objects.get_or_create(user=user, course_id=course_id)
                for thread_id, timestamp in read_state.get(
                    "last_read_times", {}
                ).items():
                    mongo_content = MongoContent.objects.filter(
                        mongo_id=thread_id
                    ).first()
                    thread = mongo_content and mongo_content.content

                    # For older courses using cs_comment_service, the thread may be None
                    # because cs_comment_service retains read_states for deleted threads
                    # in the users collection in MongoDB. As a result, MongoContent won't
                    # have a thread with deleted thread_id from read_states.
                    if not thread:
                        continue
                    existing_read_time = LastReadTime.objects.filter(
                        read_state=rs, comment_thread=thread
                    ).first()
                    if not existing_read_time:
                        LastReadTime.objects.create(
                            read_state=rs,
                            comment_thread=thread,
                            timestamp=make_aware(timestamp),
                        )
                    else:
                        existing_read_time.timestamp = make_aware(timestamp)
                        existing_read_time.save()


def delete_course_data(
    db: Database[dict[str, Any]],
    course_id: str,
    dry_run: bool,
    stdout: OutputWrapper,
) -> None:
    """Delete content (threads and comments)."""
    contents = db.contents.find({"course_id": course_id})
    for content in contents:
        subscriptions = (
            db.subscriptions.delete_many({"source_id": str(content["_id"])})
            if not dry_run
            else None
        )
    stdout.write(
        f"Subscription documents to be deleted: {subscriptions.deleted_count if subscriptions else 'N/A (dry run)'}"
    )

    content_result = (
        db.contents.delete_many({"course_id": course_id}) if not dry_run else None
    )
    stdout.write(
        f"Content documents to be deleted: {content_result.deleted_count if content_result else 'N/A (dry run)'}"
    )
    user_ids = db.users.distinct("_id", {"course_stats.course_id": course_id})

    for user_id in user_ids:
        if not dry_run:
            db.users.update_one(
                {"_id": user_id},
                {
                    "$pull": {
                        "course_stats": {"course_id": course_id},
                        "read_states": {"course_id": course_id},
                    }
                },
            )
            stdout.write(f"Updated user data for user ID: {user_id}")

    if not dry_run:
        db.users.update_many(
            {},
            {
                "$pull": {
                    "course_stats": {"course_id": course_id},
                    "read_states": {"course_id": course_id},
                }
            },
        )
        stdout.write("Cleaned up users collection")


def log_deletion(
    collection_name: str,
    result: Collection[dict[str, Any]],
    stdout: OutputWrapper,
) -> None:
    """Log the deletion of a collection."""
    stdout.write(f"Deleted {result.deleted_count} documents from {collection_name}")


# pylint: disable=import-error,import-outside-toplevel
def enable_mysql_backend_for_course(course_id: str) -> None:
    """Enable MySQL backend waffle flag for a course."""
    from opaque_keys.edx.keys import CourseKey

    from openedx.core.djangoapps.waffle_utils.models import (  # type: ignore[import-not-found]
        WaffleFlagCourseOverrideModel,
    )

    from forum.toggles import ENABLE_MYSQL_BACKEND

    course_key = CourseKey.from_string(course_id)
    WaffleFlagCourseOverrideModel.objects.create(
        course_id=course_key, waffle_flag=ENABLE_MYSQL_BACKEND.name, enabled=True
    )



"""Users Class for mongo backend."""

from typing import Any, Optional

from forum.backends.mongodb.base_model import MongoBaseModel


class Users(MongoBaseModel):
    """
    Users class for cs_comments_service user model
    """

    COLLECTION_NAME: str = "users"

    def get(self, _id: str) -> Optional[dict[str, Any]]:
        """
        Get the user based on the id
        """
        return self._collection.find_one({"_id": _id})

    def insert(
        self,
        external_id: str,
        username: Optional[str] = None,
        email: Optional[str] = None,
        default_sort_key: Optional[str] = "date",
        read_states: Optional[list[dict[str, Any]]] = None,
        course_stats: Optional[list[dict[str, Any]]] = None,
    ) -> str:
        """
        Inserts a new user document into the database.

        Args:
            external_id: The external ID of the user.
            username: The username of the user.
            email: The email of the user.
            default_sort_key: The default sort key for the user.
            read_states: The read states of the user.
            course_stats: The course statistics of the user.

        Returns:
            The ID of the inserted document.

        """
        user_data: dict[str, Any] = {
            "_id": external_id,
            "external_id": external_id,
            "username": username,
            "email": email,
            "default_sort_key": default_sort_key,
            "read_states": read_states,
            "course_stats": course_stats,
        }
        insert_data = {k: v for k, v in user_data.items() if v is not None}
        result = self._collection.insert_one(insert_data)
        return str(result.inserted_id)

    def delete(self, _id: Any) -> int:
        """
        Deletes a user document from the database based on the id.

        Args:
            _id: The ID of the user.

        Returns:
            The number of documents deleted.

        """
        result = self._collection.delete_one({"_id": _id})
        return result.deleted_count

    def update(
        self,
        external_id: str,
        username: Optional[str] = None,
        email: Optional[str] = None,
        default_sort_key: Optional[str] = None,
        read_states: Optional[list[dict[str, Any]]] = None,
        course_stats: Optional[list[dict[str, Any]]] = None,
    ) -> int:
        """
        Updates a user document in the database based on the external_id.

        Args:
            external_id: The external ID of the user.
            **kwargs: Keyword arguments to update the user document.
            Supported keys:
                - username: The new username of the user.
                - email: The new email of the user.
                - default_sort_key: The new default sort key for the user.
                - read_states: The new read states of the user.
                - course_stats: The new course statistics of the user.

        Returns:
            The number of documents modified.

        """
        fields = [
            ("username", username),
            ("email", email),
            ("default_sort_key", default_sort_key),
            ("read_states", read_states),
            ("course_stats", course_stats),
        ]
        update_data: dict[str, Any] = {
            field: value for field, value in fields if value is not None
        }

        result = self._collection.update_one(
            {"external_id": external_id},
            {"$set": update_data},
        )
        return result.modified_count

    def delete_read_state_by_thread_id(self, thread_id: str) -> None:
        """Delete read state from users based on thread_id."""
        users = self.get_list(
            **{
                "read_states.last_read_times": {"$exists": True},
                f"read_states.last_read_times.{thread_id}": {"$exists": True},
            }
        )
        for user in list(users):
            updated_read_states = []
            for read_state in user.get("read_states", []):
                if read_state["last_read_times"].get(thread_id):
                    del read_state["last_read_times"][thread_id]
                updated_read_states.append(read_state)
            self._collection.update_one(
                {"_id": user["_id"]}, {"$set": {"read_states": updated_read_states}}
            )



"""
Search API Views
"""

from typing import Any

from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from forum.api.search import search_threads
from forum.constants import FORUM_DEFAULT_PAGE, FORUM_DEFAULT_PER_PAGE
from forum.utils import get_commentable_ids_from_params, get_group_ids_from_params


class SearchThreadsView(APIView):
    """
    A view that handles the search and retrieval of threads based on query parameters.

    This view provides a `GET` endpoint that allows searching for threads with various filtering,
    sorting, and pagination options. It also supports suggesting corrected search text if no results
    are found with the initial query.
    """

    permission_classes = (AllowAny,)

    def _validate_and_extract_params(self, request: Request) -> dict[str, Any]:
        """
        Validate and extract query parameters from the request.
        """
        data = request.query_params
        params: dict[str, Any] = {}

        # Required parameters
        text = data.get("text")
        if not text:
            raise ValueError("text is required")
        params["text"] = text

        # Sort key validation
        VALID_SORT_KEYS = ("activity", "comments", "date", "votes")
        sort_key = data.get("sort_key", "date")
        if sort_key not in VALID_SORT_KEYS:
            raise ValueError("invalid sort_key")
        params["sort_key"] = sort_key

        # Pagination handling
        page = data.get("page", FORUM_DEFAULT_PAGE)
        try:
            params["page"] = int(page)
        except ValueError as exc:
            raise ValueError("Invalid page value.") from exc

        per_page = data.get("per_page", FORUM_DEFAULT_PER_PAGE)
        try:
            params["per_page"] = int(per_page)
        except ValueError as exc:
            raise ValueError("Invalid per_page value.") from exc

        # Optional parameters with default values and type conversion
        params["context"] = data.get("context", "course")
        params["user_id"] = data.get("user_id", "")
        params["course_id"] = data.get("course_id", "")
        params["author_id"] = data.get("author_id")
        params["thread_type"] = data.get("thread_type")
        params["flagged"] = data.get("flagged", "false").lower() == "true"
        params["unread"] = data.get("unread", "false").lower() == "true"
        params["unanswered"] = data.get("unanswered", "false").lower() == "true"
        params["unresponded"] = data.get("unresponded", "false").lower() == "true"
        params["count_flagged"] = data.get("count_flagged", "false").lower() == "true"

        # Group IDs extraction
        params["group_ids"] = get_group_ids_from_params(data)

        params["commentable_ids"] = get_commentable_ids_from_params(data)

        return params

    def get(self, request: Request) -> Response:
        """
        Handle GET requests to search for threads based on query parameters.

        Args:
            request (Request): The Django request object containing query parameters.

        Returns:
            Response: A JSON response containing the search results, corrected text (if any), and total results.
        """

        try:
            params: dict[str, Any] = self._validate_and_extract_params(request)
        except ValueError as error:
            return Response({"error": str(error)}, status=status.HTTP_400_BAD_REQUEST)

        search_threads_data = search_threads(**params)

        return Response(search_threads_data)



# Generated by Django 4.2.17 on 2025-03-04 06:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0002_alter_readstate_course_id_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="commentthread",
            name="title",
            field=models.CharField(max_length=1024),
        ),
    ]



"""
Database models for forum.
"""

from abc import ABC
from typing import Any, Optional

from bson import ObjectId
from pymongo.collection import Collection as PymongoCollection
from pymongo.command_cursor import CommandCursor
from pymongo.cursor import Cursor

from forum.mongo import Database, get_database

Collection = PymongoCollection[dict[str, Any]]


class MongoBaseModel(ABC):
    """Abstract Class for Mongo model implementation"""

    MONGODB_DATABASE: Optional[Database] = None
    COLLECTION_NAME: str = "default"
    index_name: str = "default"

    @property
    def _collection(self) -> Collection:
        """Return the MongoDB collection for the model."""
        return self.__get_database()[self.COLLECTION_NAME]

    @classmethod
    def __get_database(cls) -> Database:
        """Get or create static class database."""
        if cls.MONGODB_DATABASE is None:
            cls.MONGODB_DATABASE = get_database()
        return cls.MONGODB_DATABASE

    def override_query(self, query: dict[str, Any]) -> dict[str, Any]:
        """Override Query"""
        return query

    def get(self, _id: str) -> Optional[dict[str, Any]]:
        """Get a document by ID."""
        return self._collection.find_one({"_id": ObjectId(_id)})

    def get_list(self, **kwargs: Any) -> Cursor[dict[str, Any]]:
        """Get a list of all documents filtered by kwargs."""
        return self._collection.find(kwargs)

    def delete(self, _id: str) -> int:
        """
        Delete a document from the database based on the ID.

        Args:
            _id: The ID of the document.

        Returns:
            The number of documents deleted.
        """
        result = self._collection.delete_one({"_id": ObjectId(_id)})
        return result.deleted_count

    def find(self, query: dict[str, Any]) -> Cursor[dict[str, Any]]:
        """
        Run a raw MongoDB query.

        Args:
            query: The MongoDB query.

        Returns:
            A cursor with the query results.
        """
        query = self.override_query(query)
        return self._collection.find(query)

    def find_one(self, query: dict[str, Any]) -> Optional[dict[str, Any]]:
        """
        Run a raw MongoDB query to find a single document.

        Args:
            query: The MongoDB query.

        Returns:
            The first document matching the query, or None if no document matches.
        """
        query = self.override_query(query)
        return self._collection.find_one(query)

    def aggregate(
        self, pipeline: list[dict[str, Any]]
    ) -> CommandCursor[dict[str, Any]]:
        """
        Run a MongoDB aggregation pipeline.

        Args:
            pipeline: The aggregation pipeline.

        Returns:
            A command cursor with the aggregation results.
        """
        return self._collection.aggregate(pipeline)

    def count_documents(self, query: dict[str, Any]) -> int:
        """
        Count the number of documents matching a query.

        Args:
            query: The MongoDB query.

        Returns:
            The count of documents matching the query.
        """
        query = self.override_query(query)
        return self._collection.count_documents(query)

    def distinct(self, field: str, query: dict[str, Any]) -> list[Any]:
        """
        Run a MongoDB distinct query.

        Args:
            field: The field for which to return distinct values.
            query: The MongoDB query.

        Returns:
            A list of distinct values for the specified field.
        """
        query = self.override_query(query)
        return self._collection.distinct(field, query)
